# Copyright (C) 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class ParameterDefinition:
    def afun(self, lab="", **kwargs):
        """Specifies units for angular functions in parameter expressions.

        APDL Command: ``*AFUN``

        Parameters
        ----------
        lab
            Specifies the units to be used:

            Use radians for input and output of parameter angular functions (default). - Use degrees for input and output of parameter angular functions.

        Notes
        -----
        Only the SIN, COS, TAN, ASIN, ACOS, ATAN, ATAN2, ANGLEK, and ANGLEN
        functions ``[*SET, *VFUN]`` are affected by this command.
        """
        command = f"*AFUN,{lab}"
        return self.run(command, **kwargs)

    def dim(
        self,
        par="",
        type_="",
        imax="",
        jmax="",
        kmax="",
        var1="",
        var2="",
        var3="",
        csysid="",
        **kwargs,
    ):
        """Defines an array parameter and its dimensions.

        APDL Command: ``*DIM``

        Parameters
        ----------
        par
            Name of parameter to be dimensioned.  See ``*SET`` for name
            restrictions.

        type\\_
            Array type:

            Arrays are similar to standard FORTRAN arrays (indices are
            integers) (default).  Index numbers for the rows, columns,
            and planes are sequential values beginning with one. Used
            for 1-, 2-, or 3-D arrays. - Same as ARRAY, but used to
            specify 4-D arrays.

            Same as ARRAY, but used to specify 5-D arrays. - Array
            entries are character strings (up to 8 characters each).
            Index numbers for rows, columns, and planes are sequential
            values beginning with one.

            Array indices are real (non-integer) numbers which must be
            defined when filling the table.  Index numbers for the
            rows and columns are stored in the zero column and row
            "array elements" and are initially assigned a near-zero
            value.  Index numbers must be in ascending order and are
            used only for retrieving an array element.  When
            retrieving an array element with a real index that does
            not match a specified index, linear interpolation is done
            among the nearest indices and the corresponding array
            element values [``*SET``]. Used for 1-, 2-, or 3-D tables. -
            Same as TABLE, but used to specify 4-D tables.

            Same as TABLE, but used to specify 5-D tables. - Array
            entries are character strings (up to IMAX each). Index
            numbers for columns and planes are sequential values
            beginning with 1. Row index is character position in
            string.

        imax
            Extent of first dimension (row). (For Type = STRING, IMAX is
            rounded up to the next multiple of eight and has a limit of 248).
            Defaults to 1.

        jmax
            Extent of second dimension (column).  Defaults to 1.

        kmax
            Extent of third dimension (plane).  Defaults to 1.

        var1
            Variable name corresponding to the first dimension (row) for Type =
            TABLE.  Defaults to Row.

        var2
            Variable name corresponding to the second dimension (column) for
            Type = TABLE.  Defaults to Column.

        var3
            Variable name corresponding to the third dimension (plane) for Type
            = TABLE.  Defaults to Plane.

        csysid
            An integer corresponding to the coordinate system ID Number.

        Notes
        -----
        Up to three dimensions (row, column, and plane) may be defined using
        ARRAY and TABLE.  Use ARR4, ARR5, TAB4, and TAB5 to define up to five
        dimensions (row, column, plane, book, and shelf). An index number is
        associated with each row, column, and plane.  For array and table type
        parameters, element values are initialized to zero.  For character and
        string parameters, element values are initialized to (blank).  A
        defined parameter must be deleted [``*SET``] before its dimensions can be
        changed.  Scalar (single valued) parameters should not be dimensioned.
        ``*DIM,A,,3`` defines a vector array with elements A(1), A(2), and A(3).
        ``*DIM,B,,2,3`` defines a 2x3 array with elements B(1,1), B(2,1), B(1,2),
        B(2,2), B(1,3), and B(2,3).  Use ``*STATUS,Par`` to display elements of
        array Par. You can write formatted data files (tabular formatting) from
        data held in arrays through the ``*VWRITE`` command.

        If you use table parameters to define boundary conditions, then Var1,
        Var2, and/or Var3 can either specify a primary variable (listed in
        Table: 130:: ``*DIM`` - Primary Variables) or can be an independent
        parameter.  If specifying an independent parameter, then you must
        define an additional table for the independent parameter.  The
        additional table must have the same name as the independent parameter
        and may be a function of one or more primary variables or another
        independent parameter.  All independent parameters must relate to a
        primary variable.

        Tabular load arrays can be defined in both global Cartesian (default)
        or local (see below) coordinate systems by specifying CSYSID, as
        defined in LOCAL. For batch operations, you must specify your
        coordinate system first.

        The following constraints apply when you specify a local coordinate
        system for your tabular loads:

        If you are specifying a 4- or 5-D array or table, four additional
        fields (LMAX, MMAX, Var4, and Var5) are available. Thus, for a 4-D
        table, the command syntax would be:

        For a 5-D table, the command syntax would be:

        You cannot create or edit 4- or 5-D arrays or tables using the GUI.

        See Array Parameters for a detailed discussion on and examples for
        using array parameters.

        Table: 130:: : ``*DIM`` - Primary Variables

        Specify PRESSURE as the independent variable (not PRES).

        The X, Y, and Z coordinate locations listed above are valid in global
        Cartesian, or local (Cartesian, cylindrical and spherical) coordinate
        systems. The VELOCITY label is applicable only to the calculated fluid
        velocity in element FLUID116.

        When using PRESSURE as a primary variable, the underlying element must
        have the pressure DOF associated with it, or it must be a supported
        contact element.

        The gap/penetration label (GAP) is only used for defining certain
        contact element real constants.

        The frequency label (FREQ) is valid for harmonic analyses only.

        The OMEGS, ECCENT, and THETA primary variables only apply to the
        COMBI214 element. The amplitude of the rotational velocity (OMEGS) is
        an absolute value, so only positive values of OMEGS are valid. The
        eccentricity (ECCENT) and phase shift (THETA) labels are only valid for
        nonlinear analyses.

        If you use table parameters to define boundary conditions, the table
        names (Par) must not exceed 32 characters.

        In thermal analyses, if you apply tabular loads as a function of
        temperature but the rest of the model is linear (e.g., includes no
        temperature-dependent material properties or radiation ), you should
        turn on Newton-Raphson iterations (NROPT,FULL) to evaluate the
        temperature-dependent tabular boundary conditions correctly.

        This command is valid in any processor.
        """
        command = (
            f"*DIM,{par},{type_},{imax},{jmax},{kmax},{var1},{var2},{var3},{csysid}"
        )
        return self.run(command, **kwargs)

    def get(
        self,
        par="",
        entity="",
        entnum="",
        item1="",
        it1num="",
        item2="",
        it2num="",
        **kwargs,
    ):
        """Retrieves a value and stores it as a scalar parameter or part of an array parameter.

        APDL Command: ``*GET``

        See the full MADPL command at `*GET
        <https://www.mm.bme.hu/~gyebro/files/ans_help_v182/ans_cmd/Hlp_C_GET.html>`_

        GET retrieves a value for a specified item and stores the
        value as a scalar parameter, or as a value in a user-named
        array parameter. An item is identified by various keyword,
        label, and number combinations.  Usage is similar to the SET
        command except that the parameter values are retrieved from
        previously input or calculated results. For example,
        ``GET,A,ELEM,5,CENT,X`` returns the centroid x-location of element
        5 and stores the result as parameter A. GET command
        operations, along with the associated Get functions return
        values in the active coordinate system unless stated
        otherwise. A Get function is an alternative in- line function
        that can be used to retrieve a value instead of the GET
        command (see Using In-line Get Functions for more
        information).
        Both GET and VGET retrieve information from the active data
        stored in memory. The database is often the source, and
        sometimes the information is retrieved from common memory
        blocks that the program uses to manipulate
        information. Although POST1 and POST26 operations use a .rst
        file, GET data is accessed from the database or from the
        common blocks. Get operations do not access the .rst file
        directly. For repeated gets of sequential items, such as from
        a series of elements, see the VGET command.
        Most items are stored in the database after they are
        calculated and are available anytime thereafter. Items are
        grouped according to where they are usually first defined or
        calculated. Preprocessing data will often not reflect the
        calculated values generated from section data. Do not use GET
        to obtain data from elements that use calculated section data,
        such as beams or shells. Most of the general items listed
        below are available from all modules.

        Parameters
        ----------
        par : str, optional
            The name of the resulting parameter. See \\*SET for name
            restrictions.

        entity
            Entity keyword. Valid keywords are NODE, ELEM, KP, LINE,
            AREA, VOLU, PDS, etc., as shown for Entity = in the tables
            below.

        entnum
            The number or label for the entity (as shown for ENTNUM =
            in the tables below). In some cases, a zero (or blank)
            ENTNUM represents all entities of the set.

        item1
            The name of a particular item for the given entity.

        it1num
            The number (or label) for the specified Item1 (if
            any). Valid IT1NUM values are as shown in the IT1NUM
            columns of the tables below. Some Item1 labels do not
            require an IT1NUM value.

        item2, it2num
            A second set of item labels and numbers to further qualify
            the item for which data are to be retrieved. Most items do
            not require this level of information.

        item3
            A third set of item labels to further qualify
            the item for which data are to be retrieved. Almost all items do
            not require this level of information.

        item3 : str, optional
            A third set of item labels and numbers to further qualify the item
            for which data are to be retrieved. Most items do not require this
            level of information.

        it3num : str, int, optional
            The number (or label) for the specified ``item3`` (if
            any). Some ``item3`` labels do not require an ``it3num``
            value.

        item4 : str, optional
            A fourth set of item labels and numbers to further qualify the item
            for which data are to be retrieved. Most items do not require this level of information.

        it4num : str, int, optional
            The number (or label) for the specified ``item4`` (if
            any). Some ``item4`` labels do not require an ``it4num``
            value.

        Returns
        -------
        float
            Floating point value of the parameter.

        Examples
        --------
        Retrieve the number of nodes

        >>> value = mapdl.get('val', 'node', '', 'count')
        >>> value
        3003

        Retrieve the number of nodes using keywords.  Note that the
        parameter name is optional.

        >>> value = mapdl.get(entity='node', item1='count')
        >>> value
        3003

        """
        command = f"*GET,{par},{entity},{entnum},{item1},{it1num},{item2},{it2num}"
        return self.run(command, **kwargs)

    def inquire(self, strarray="", func="", arg1="", arg2=""):
        """Returns system information.

        By default, with no arguments, it returns the working directory.

        >>> mapdl.inquire()
        C:\\Users\\user\\AppData\\Local\\Temp\\ansys_nynvxsaooh

        Parameters
        ----------
        strarray : str, optional
            Name of the string array or parameter that will hold the returned values.
            Normally, if used in a python script you should just work with the
            return value from this method.
        func : str, optional
           Specifies the type of system information returned.  See the
           notes section for more information.
        arg1 : str, optional
            First argument. See notes for ``arg1`` definition.
        arg2 : str, optional
            Second argument. See notes for ``arg1`` definition.

        Returns
        -------
        str
            Value of the inquired item.

        Notes
        -----
        The ``/INQUIRE`` command is valid in any processor.

        .. warning::
           Take note that from version 0.60.4 and later, the command behaviour
           has been changed.
           Previously, the ``StrArray`` argument was omitted. For example:
           >>> mapdl.inquire('DIRECTORY')
           C:\\Users\\user\\AppData\\Local\\Temp\\ansys_nynvxsaooh

           Now this will raise an exception.
           The default behaviour now, requires to input ``StrArray``:
           >>> mapdl.inquire('', 'DIRECTORY')
           C:\\Users\\user\\AppData\\Local\\Temp\\ansys_nynvxsaooh

        **GENERAL FUNC OPTIONS**

        - ``LOGIN`` - Returns the pathname of the login directory on Linux
          systems or the pathname of the default directory (including
          drive letter) on Windows systems.
        - ``DOCU`` - Pathname of the ANSYS documentation directory.
        - ``APDL`` - Pathname of the ANSYS APDL directory.
        - ``PROG`` - Pathname of the ANSYS executable directory.
        - ``AUTH`` - Pathname of the directory in which the license file resides.
        - ``USER`` - Name of the user currently logged-in.
        - ``DIRECTORY`` - Pathname of the current directory.
        - ``JOBNAME`` - Current Jobname.
        - ``RSTDIR`` - Result file directory.
        - ``RSTFILE`` - Result file name.
        - ``RSTEXT`` - Result file extension.
        - ``OUTPUT`` - Current output file name.


        **RETURNING THE VALUE OF AN ENVIRONMENT VARIABLE TO A PARAMETER**

        If ``FUNC=ENV``, the command format is ``/INQUIRE,StrArray,ENV,ENVNAME,Substring``.
        In this instance, ENV specifies that the command should return the
        value of an environment variable.
        The following defines the remaining fields:

        Envname:
            Specifies the name of the environment variable.

        Substring:
            If ``Substring = 1``, the first substring (up to the first colon (:)) is returned.
            If ``Substring = 2``, the second substring is returned, etc. For Windows platforms,
            the separating character is semicolon (;).
            If this argument is either blank or 0, the entire value of the environment
            variable is returned.


        **RETURNING THE VALUE OF A TITLE TO A PARAMETER**

        If ``FUNC = TITLE``, the command format is ``/INQUIRE,StrArray,TITLE,Title_num``.
        In this context, the value of Title_num can be blank or ``1`` through ``5``. If the
        value is ``1`` or blank, the title is returned. If the value is ``2`` through ``5``,
        a corresponding subtitle is returned (``2`` denoting the first subtitle, and so on).


        **RETURNING INFORMATION ABOUT A FILE TO A PARAMETER**

        The ``/INQUIRE`` command can also return information about specified files
        within the file system.
        For these capabilities, the format is ``/INQUIRE,Parameter,FUNC,Fname, Ext, --``.
        The following defines the fields:

        Parameter:
            Name of the parameter that will hold the returned values.

        Func:
            Specifies the type of file information returned:

            EXIST:
                Returns a ``1`` if the specified file exists, and ``0`` if it does not.

            DATE:
                Returns the date stamp of the specified file in the format ``*yyyymmdd.hhmmss*``.

            SIZE:
                Returns the size of the specified file in MB.

            WRITE:
                Returns the status of the write attribute. A ``0`` denotes no write permission while a ``1`` denotes
                write permission.

            READ:
                Returns the status of the read attribute. A ``0`` denotes no read permission while a ``1`` denotes read
                permission.

            EXEC:
                Returns the status of the execute attribute (this has meaning only on Linux). A ``0`` denotes no
                execute permission while a ``1`` denotes execute permission.

            LINES:
                Returns the number of lines in an ASCII file.

        Fname:
            File name and directory path (248 characters maximum, including the characters needed for the
            directory path). An unspecified directory path defaults to the working directory; in this case, you
            can use all 248 characters for the file name.

        Ext:
            Filename extension (eight-character maximum).

        Examples
        --------
        Return the MAPDL working directory
        >>> mapdl.inquire('', 'DIRECTORY')
        C:\\Users\\gayuso\\AppData\\Local\\Temp\\ansys_nynvxsaooh

        Or

        >>> mapdl.inquire()
        C:\\Users\\gayuso\\AppData\\Local\\Temp\\ansys_nynvxsaooh

        Return the job name

        >>> mapdl.inquire('', 'JOBNAME')
        file

        Return the result file name

        >>> mapdl.inquire('', 'RSTFILE')
        'file.rst'
        """
        return self.run(f"/INQUIRE,{strarray},{func},{arg1},{arg2}")

    def parres(self, lab="", fname="", ext="", **kwargs):
        """Reads parameters from a file.

        APDL Command: PARRES

        Reads parameters from a coded file.  The parameter file may
        have been written with the PARSAV command.  The parameters
        read may replace or change the current parameter set.

        This command is valid in any processor.

        Parameters
        ----------
        lab
            Read operation.

            NEW - Replace current parameter set with these parameters (default).

            CHANGE - Extend current parameter set with these
            parameters, replacing any that already exist.

        fname
            File name and directory path (248 characters maximum,
            including the characters needed for the directory path).
            An unspecified directory path defaults to the working
            directory; in this case, you can use all 248 characters
            for the file name.

            The file name defaults to Jobname.

        ext
            Filename extension (eight-character maximum).  The
            extension defaults to PARM if Fname is blank.

        Examples
        --------
        Read a local parameter file.

        >>> mapdl.parres('parm.PARM')

        """
        return self.run(f"PARRES, {lab}, {fname}, {ext}")

    def parsav(self, lab="", fname="", ext="", **kwargs):
        """Writes parameters to a file.

        APDL Command: PARSAV

        Parameters
        ----------
        lab
            Write operation:

            - ``'SCALAR'`` : Write only scalar parameters (default).
            - ``'ALL'`` : Write scalar and array parameters.
              Parameters may be numeric or alphanumeric.

        fname
            File name and directory path (248 characters maximum,
            including the characters needed for the directory path).
            An unspecified directory path defaults to the working
            directory; in this case, you can use all 248 characters
            for the file name.

        ext
            Filename extension (eight-character maximum).

        Notes
        -----
        Writes the current parameters to a coded file.  Previous
        parameters on this file, if any, will be overwritten.  The
        parameter file may be read with the PARRES command.

        PARSAV/PARRES operations truncate some long decimal strings,
        and can cause differing values in your solution data when
        other operations are performed. A good practice is to limit
        the number of decimal places you will use before and after
        these operations.

        This command is valid in any processor.
        """
        command = f"PARSAV,{lab},{fname},{ext}"
        return self.run(command, **kwargs)

    def starset(
        self,
        par="",
        value="",
        val2="",
        val3="",
        val4="",
        val5="",
        val6="",
        val7="",
        val8="",
        val9="",
        val10="",
        **kwargs,
    ):
        """Assigns values to user-named parameters.

        APDL Command: ``*SET``

        Parameters
        ----------
        par
            An alphanumeric name used to identify this parameter.  Par may be
            up to 32 characters, beginning with a letter and containing only
            letters, numbers, and underscores.  Examples:  ABC   A3X   TOP_END.
            ANSYS command names, function names, label names, component and
            assembly names, etc., should not be used.  Parameter names
            beginning with an underscore (e.g.,  _LOOP) are reserved for use by
            ANSYS and should be avoided.  Parameter names ending in an
            underscore are not listed by the ``*STATUS`` command.  Array parameter
            names must be followed by a subscript, and the entire expression
            must be 32 characters or less.  Examples:  A(1,1)   NEW_VAL(3,2,5)
            RESULT(1000).  There is no character parameter substitution for the
            Par field. Table parameters that are used in command fields (where
            constant values are normally given) are limited to 32 characters.

        value
            Numerical value or alphanumeric character string (up to 32
            characters enclosed in single quotes) to be assigned to this
            parameter.  Examples:  A(1,3)=7.4 B='ABC3'.  May also be a
            parameter or a parametric expression.  Examples:  C=A(1,3)
            A(2,2)=(C+4)/2.  If blank, delete this parameter.  Example:  A=
            deletes parameter A.

        val2, val3, val4, val5, val6, val7, val8, val9, val10
            If Par is an array parameter, values VAL2 through VAL10 (up to the
            last nonblank value) are sequentially assigned to the succeeding
            array elements of the column.  Example:  ``*SET,A(1,4),10,11`` assigns
            ``A(1,4)=10, A(2,4)=11``.  ``*SET,B(2,3),'file10','file11'`` assigns
            ``B(2,3)='file10', B(3,3)='file11'``.

        Notes
        -----
        Assigns values to user-named parameters that may be substituted later
        in the run.  The equivalent (and recommended) format is

        ``Par = VALUE,VAL2,VAL3, . . . , VAL10``

        which may be used in place of  ``*SET,Par, : ...`` for convenience.

        This command is valid in any processor.

        Parameters (numeric or character) may be scalars (single valued) or
        arrays (multiple valued in one, two, or three dimensions). An unlimited
        number of parameter names may be defined in any ANSYS run. For very
        large numbers of parameters, it is most efficient to define them in
        alphabetical order.

        Parameter values may be redefined at any time.  Array parameters may
        also be assigned values within a do-loop [``*DO``] for convenience.
        Internally programmed do-loop commands are also available with the ``*VXX``
        commands (``*VFILL``).  Parameter values (except for parameters ending in
        an underscore) may be listed with the  ``*STATUS`` command, displayed with
        the ``*VPLOT``  command (numeric parameters only), and modified with the
        ``*VEDIT`` command (numeric parameters only).

        Older ANSYS-supplied macro files may use parameter names that do not
        begin with an underscore. Using these macros embedded in your own
        macros may cause conflicts if the same parameter names are used.

        Parameters can also be resolved in comments created by the /COM command
        (see /COM for complete documentation). A parameter can be deleted by
        redefining it with a blank  VALUE.  If the parameter is an array, the
        entire array is deleted.  Parameters may also be defined by a response
        to a query with the  ``*ASK`` command or from an "ANSYS-supplied" value
        with the ``*GET`` command.

        Array parameters must be dimensioned  [``*DIM``] before being assigned
        values unless they are the result of an array operation or defined
        using the implied loop convention. Scalar parameters that are not
        defined are initialized to a "near" zero value.  Numeric array
        parameters are initialized to zero when dimensioned, and character
        array parameters are initialized to blank.  An existing array parameter
        must be deleted before it can be redimensioned.  Array parameter names
        must be followed by a subscript list (enclosed in parentheses)
        identifying the element of the array.  The subscript list may have one,
        two, or three values (separated by commas).  Typical array parameter
        elements are  A(1,1), NEW_VAL(3,2,5), RESULT(1000).  Subscripts for
        defining an array element must be integers (or parameter expressions
        that evaluate to integers).  Non-integer values are rounded to the
        nearest integer value.  All array parameters are stored as 3-D arrays
        with the unspecified dimensions set to 1.  For example, the 4th array
        element of a 1-dimensional array, A(4), is stored as array element
        A(4,1,1).  Arrays are patterned after standard FORTRAN conventions.

        If the parameter name Par is input in a numeric argument of a command,
        the numeric value of the parameter (as assigned with ``*SET``, ``*GET``, =,
        etc.) is substituted into the command at that point.  Substitution
        occurs only if the parameter name is used between blanks, commas,
        parentheses, or arithmetic operators (or any combination) in a numeric
        argument.  Substitution can be prevented by enclosing the parameter
        name Par within single quotes ( ' ), if the parameter is alone in the
        argument; if the parameter is part of an arithmetic expression, the
        entire expression must be enclosed within single quotes to prevent
        substitution.  In either case the character string will be used instead
        of the numeric value (and the string will be taken as 0.0 if it is in a
        numeric argument).

        A forced substitution is available in the text fields of the /TITLE,
        /STITLE,  /TLABEL, /AN3D, /SYP (ARG1--ARG8), and ``*ABBR``  commands by
        enclosing the parameter within percent (%) signs.  Also, parameter
        substitution may be forced within the file name or extension fields of
        commands having these fields by enclosing the parameter within percent
        (%) signs.  Array parameters  [``*DIM``] must include a subscript (within
        parentheses) to identify the array element whose value is to be
        substituted, such as A(1,3).  Out-of-range subscripts result in an
        error message.  Non-integer subscripts are allowed when identifying a
        TABLE array element for substitution.  A proportional linear
        interpolation of values among the nearest array elements is performed
        before substitution.  Interpolation is done in all three dimensions.

        Note:: : Interpolation is based upon the assigned index numbers which
        must be defined when the table is filled [``*DIM``].

        Most alphanumeric arguments permit the use of character
        parameter substitution.  When the parameter name Par input,
        the alphanumeric value of the parameter is substituted into
        the command at that point.  Substitution can be suppressed by
        enclosing the parameter name within single quotes ( ' ).
        Forced substitution is available in some fields by enclosing
        the parameter name within percent (%) signs.  Valid forced
        substitution fields include command name fields, Fname
        (filename) or Ext (extension) arguments, ``*ABBR`` command
        (Abbr arguments), /TITLE and /STITLE commands (Title argument)
        and /TLABEL command (Text argument).  Character parameter
        substitution is also available in the ``*ASK``, /AN3D,
        ``*CFWRITE``, ``*IF``, ``*ELSEIF``, ``*MSG``, ``*SET``,
        ``*USE``, ``*VREAD``, and ``*VWRITE`` commands.  Character array
        parameters must include a subscript (within parentheses) to
        identify the array element whose value is to be substituted.

        If a parameter operation expression is input in a numeric argument, the
        numeric value of the expression is substituted into the command at that
        point.  Allowable operation expressions are of the form

        ``E1oE2oE3: ...oE10``

        where E1, E2, etc. are expressions connected by operators (o).  The
        allowable operations (o) are

        ``+ - * / ** < >``

        For example, ``A+B**C/D*E`` is a valid operation expression.  The ``*``
        represents multiplication and the ``**`` represents exponentiation.
        """
        command = f"*SET,{par},{value},{val2},{val3},{val4},{val5},{val6},{val7},{val8},{val9},{val10}"
        return self.run(command, **kwargs)

    def starvget(
        self,
        parr="",
        entity="",
        entnum="",
        item1="",
        it1num="",
        item2="",
        it2num="",
        kloop="",
        **kwargs,
    ):
        """Retrieves values and stores them into an array parameter.

        APDL Command: ``*VGET``

        Parameters
        ----------
        parr
            The name of the resulting vector array parameter.  See ``*SET`` for
            name restrictions.

        entity
            Entity keyword.  Valid keywords are NODE, ELEM, KP, LINE, AREA,
            VOLU, etc. as shown for Entity = in the tables below.

        entnum
            The number of the entity (as shown for ENTNUM = in the tables
            below).

        item1
            The name of a particular item for the given entity.  Valid items
            are as shown in the Item1 columns of the tables below.

        it1num
            The number (or label) for the specified Item1 (if any).  Valid
            IT1NUM values are as shown in the IT1NUM columns of the tables
            below.  Some Item1 labels do not require an IT1NUM value.

        item2, it2num
            A second set of item labels and numbers to further qualify the item
            for which data is to be retrieved.  Most items do not require this
            level of information.

        kloop
            Field to be looped on:

            ``0`` or ``2`` : Loop on the ENTNUM field (default).

            ``3`` : Loop on the Item1 field.

            ``4`` : Loop on the IT1NUM field. Successive items are as shown with IT1NUM.

            ``5`` : Loop on the Item2 field.

            ``6`` : Loop on the IT2NUM field. Successive items are as shown with IT2NUM.

        Notes
        -----
        Retrieves values for specified items and stores the values in an output
        vector of a user-named array parameter according to:

        ``ParR = f(Entity, ENTNUM, Item1, IT1NUM, Item2, IT2NUM)``

        where (f) is the ``*GET`` function; Entity, Item1, and Item2
        are keywords; and ENTNUM, IT1NUM, and IT2NUM are numbers or
        labels corresponding to the keywords. Looping continues over
        successive entity numbers (ENTNUM) for the KLOOP default.  For
        example, ``*VGET,A(1),ELEM,5,CENT,X`` returns the centroid
        x-location of element 5 and stores the result in the first
        location of A.  Retrieving continues with element 6, 7, 8,
        etc., regardless of whether the element exists or is selected,
        until successive array locations are filled.  Use ``*VLEN`` or
        ``*VMASK`` to skip locations. Absolute values and scale
        factors may be applied to the result parameter [``*VABS``,
        ``*VFACT``].  Results may be cumulative [``*VCUM``].  See the
        ``*VOPER`` command for general details.  Results can be put
        back into an analysis by writing a file of the desired input
        commands with the ``*VWRITE`` command.  See also the ``*VPUT``
        command.

        Both ``*GET`` and ``*VGET`` retrieve information from the
        active data stored in memory. The database is often the
        source, and sometimes the information is retrieved from common
        memory blocks that ANSYS uses to manipulate
        information. Although POST1 and POST26 operations use a ``*.rst``
        file, GET data is accessed from the database or from the
        common blocks. Get operations do not access the ``*.rst`` file
        directly.

        The ``*VGET`` command retrieves both the unprocessed real and the imaginary
        parts (original and duplicate sector nodes and elements) of a cyclic
        symmetry solution.

        For each  of the sections for accessing ``*VGET`` parameters see:
        https://www.mm.bme.hu/~gyebro/files/ans_help_v182/ans_cmd/Hlp_C_VGET_st.html

        This command is valid in any processor.
        """
        command = (
            f"*VGET,{parr},{entity},{entnum},{item1},{it1num},{item2},{it2num},{kloop}"
        )
        return self.run(command, **kwargs)

    def taxis(
        self,
        parmloc="",
        naxis="",
        val1="",
        val2="",
        val3="",
        val4="",
        val5="",
        val6="",
        val7="",
        val8="",
        val9="",
        val10="",
        **kwargs,
    ):
        """Defines table index numbers.

        APDL Command: ``*TAXIS``

        Parameters
        ----------
        parmloc
            Name and starting location in the table array parameter for
            indexing. Indexing occurs along the axis defined with nAxis.

        naxis
            Axis along which indexing occurs.  Valid labels are:

            Corresponds to Row. Default. - Corresponds to Column.

            Corresponds to Plane. - Corresponds to Book.

            Corresponds to Shelf. - Lists all index numbers. Valid only if Val1 = LIST.

        val1, val2, val3, . . . , val10
            Values of the index numbers for the axis nAxis, starting from the
            table array parameter location ParmLoc. You can define up to ten
            values.

        Notes
        -----
        ``*TAXIS`` is a convenient method to define table index values. These
        values reside in the zero column, row, etc. Instead of filling values
        in these zero location spots, use the ``*TAXIS`` command.

        To list index numbers, issue ```*TAXIS,ParmLoc, nAxis, LIST,`` where
        nAxis = 1 through 5 or ALL.
        """
        command = f"*TAXIS,{parmloc},{naxis},{val1},{val2},{val3},{val4},{val5},{val6},{val7},{val8},{val9},{val10}"
        return self.run(command, **kwargs)

    def tread(self, par="", fname="", ext="", nskip="", **kwargs):
        """Reads data from an external file into a table array parameter.

        APDL Command: ``*TREAD``

        Parameters
        ----------
        par
            Table array parameter name as defined by the ``*DIM`` command.

        fname
            File name and directory path (248 characters maximum, including the
            characters needed for the directory path).  An unspecified
            directory path defaults to the working directory; in this case, you
            can use all 248 characters for the file name.

        ext
            Filename extension (eight-character maximum).

        nskip
            Number of comment lines at the beginning of the file being read
            that will be skipped during the reading.  Default = 0.

        Notes
        -----
        Use this command to read in a table of data from an external file into
        an ANSYS table array parameter.  The external file may be created using
        a text editor or by an external application or program.  The external
        file must be in tab-delimited, blank-delimited, or comma-delimited
        format to be used by ``*TREAD``. The ANSYS TABLE type array parameter must
        be defined before you can read in an external file.  See ``*DIM``  for more
        information.

        This command is not applicable to 4- or 5-D tables.
        """
        command = f"*TREAD,{par},{fname},{ext},,{nskip}"
        return self.run(command, **kwargs)

    def vfill(
        self,
        parr="",
        func="",
        con1="",
        con2="",
        con3="",
        con4="",
        con5="",
        con6="",
        con7="",
        con8="",
        con9="",
        con10="",
        **kwargs,
    ):
        """Fills an array parameter.

        APDL Command: ``*VFILL``

        Parameters
        ----------
        parr
            The name of the resulting numeric array parameter vector.  See ``*SET``
            for name restrictions.

        func
            Fill function:

            DATA - Assign specified values CON1, CON2, etc. to successive
            array elements.  Up to 10 assignments may be made at a
            time.  Any CON values after a blank CON value are
            ignored. - Assign ramp function values: ``CON1+((n-1)*CON2)``
            , where n is the loop number [``*VLEN``].  To specify a
            constant function (no ramp), set CON2 to zero.

            RAMP - Assign random number values based on a uniform
            distribution RAND(CON1,CON2), where: - Assign random
            sample of Gaussian distributions GDIS(CON1,CON2).

            RAND - Assign random number values based on a uniform
            distribution RAND(CON1,CON2), where CON1 is the lower
            bound (defaults to 0.0) and CON2 is the upper bound
            (defaults to 1.0)

            GDIS - Assign random sample of Gaussian distributions
            GDIS(CON1,CON2) where CON1 is the mean (defaults to 0.0),
            and CON2 is the standard deviation (defaults to 1.0)

            TRIA - Assigns random number values based on a triangular
            distribution TRIA(CON1,CON2,CON3) where CON1 is the lower
            bound (defaults to 0.0), CON2 is the location of the peak
            value (CON1 ≤ CON2 ≤CON3; CON2 defaults to 0 if CON1 ≤ 0 ≤
            CON3, CON1 if 0 ≤ CON1, or CON3 if CON3 ≤ 0), and CON3 is
            the upper bound (defaults to 1.0 + CON1 if CON1 ≥ 0 or 0.0
            if CON1 ≤ 0)

            BETA - Assigns random number values based on a beta
            distribution BETA(CON1,CON2,CON3,CON4) where: CON1 is the
            lower bound (defaults to 0.0), CON2 is the upper bound
            (defaults to 1.0 + CON1 if CON1 ≥ 0 or 0.0 if CON1 ≤ 0), and CON3
            and CON4 are the alpha and beta parameters, respectively,
            of the beta function. Alpha and beta must both be
            positive; they default to 1.0.

            GAMM - Assigns random number values based on a gamma
            distribution: GAMM(CON1,CON2,CON3) where: CON1 is the
            lower bound (defaults to 0.0), CON2 and CON3 are the alpha
            and beta parameters, respectively, of the gamma
            function. Alpha and beta must both be positive; they
            default to 1.0.

            RIGID - Generates the rigid body modes with respect to the
            reference point coordinates (CON1, CON2, CON3). The
            dimensions of the array parameter ParR are (dim1,dim2)
            where dim1 is the maximum node number (including internal
            nodes but excluding orientation nodes) multiplied by the
            number of degrees of freedom, and dim2 is the number of
            rigid body modes (which corresponds to the number of
            structural degrees of freedom).

            CLUSTER - Generates excitation frequencies with clustering
            option CLUSTER(CON1,CON2,CON3,CON4,%CON5%) where:

            - CON1 is the lower end of the frequency range in Hz (0 < CON1)
            - CON2 is the upper end of the frequency range in Hz (CON1 < CON2)
            - CON3 is the number of points on each side of the natural
              frequency (4 ≤ CON3 ≤ 20, defaults to 4)
            - CON4 is the constant damping ratio value or an array
              parameter (size NFR) specifying the damping ratios (if
              zero or blank, defaults to constant damping ratio of
              0.005)
            - CON5 is an array parameter (size NFR) specifying the
              natural frequencies in Hz

            The dimension of the resulting array parameter ParR is
            less than ``2+NFR*(2*CON3+1)`` where NFR is the number of
            natural frequencies defined in CON5.

        con1, con2, con3, . . . , con10
            Constants used with above functions.

        Notes
        -----
        Operates on input data and produces one output array parameter vector
        according to:

        ParR = f(CON1, CON2, : ...)

        where the functions (f) are described above. Operations use successive
        array elements [``*VLEN``, ``*VMASK``] with the default being all successive
        elements.  For example, ``*VFILL,A,RAMP,1,10`` assigns A(1) = 1.0, A(2) =
        11.0, A(3) = 21.0, etc.  ``*VFILL,B(5,1),DATA,1.5,3.0`` assigns ``B(5,1) =
        1.5 and B(6,1) = 3.0``.  Absolute values and scale factors may be applied
        to the result parameter [``*VABS``, ``*VFACT``].  Results may be cumulative
        [``*VCUM``].  See the ``*VOPER`` command for details.

        This command is valid in any processor.
        """
        command = f"*VFILL,{parr},{func},{con1},{con2},{con3},{con4},{con5},{con6},{con7},{con8},{con9},{con10}"
        return self.run(command, **kwargs)

    def vread(
        self,
        parr="",
        fname="",
        ext="",
        label="",
        n1="",
        n2="",
        n3="",
        nskip="",
        **kwargs,
    ):
        """Reads data and produces an array parameter vector or matrix.

        APDL Command: ``*VREAD``

        Parameters
        ----------
        parr
            The name of the resulting array parameter vector.  See ``*SET`` for
            name restrictions.  The parameter must exist as a dimensioned array
            [``*DIM``]. String arrays are limited to a maximum of 8 characters.

        fname
            File name and directory path (248 characters maximum, including the
            characters needed for the directory path).  An unspecified
            directory path defaults to the working directory; in this case, you
            can use all 248 characters for the file name.

        ext
            Filename extension (eight-character maximum).

        label
            Can take a value of IJK, IKJ, JIK, JKI, KIJ, KJI, or blank (IJK).

        n1, n2, n3
            Read as (((ParR (i,j,k), k = 1,n1), i = 1, n2), j = 1, n3) for
            Label = KIJ. n2 and n3 default to 1.

        nskip
            Number of lines at the beginning of the file being read that will
            be skipped during the reading.  Default = 0.

        Notes
        -----
        Reads data from a file and fills in an array parameter vector or
        matrix.  Data are read from a formatted file or, if the menu is off
        [/MENU,OFF] and Fname is blank, from the next input lines.  The format
        of the data to be read must be input immediately following the ``*VREAD``
        command.  The format specifies the number of fields to be read per
        record, the field width, and the placement of the decimal point (if
        none specified in the value).  The read operation follows the available
        FORTRAN FORMAT conventions of the system (see your system FORTRAN
        manual).  Any standard FORTRAN real format (such as (4F6.0),
        (E10.3,2X,D8.2), etc.) or alphanumeric format (A) may be used.
        Alphanumeric strings are limited to a maximum of 8 characters for any
        field (A8). For storage of string arrays greater than 8 characters, the
        ``*SREAD`` command can be used. Integer (I) and list-directed (*)
        descriptors may not be used.  The parentheses must be included in the
        format and the format must not exceed 80 characters (including
        parentheses).  The input line length is limited to 128 characters.

        A starting array element number must be defined for the result array
        parameter vector (numeric or character).  For example, entering these
        two lines:

        will read two values from each line of file ARRAYVAL and assign the
        values to A(1), A(2), A(3), etc.  Reading continues until successive
        row elements [``*VLEN``, ``*VMASK``, ``*DIM``] are filled.

        For an array parameter matrix, a starting array element row and column
        number must be defined.  For example, entering these two lines:

        will read two values from each line of file ARRAYVAL and assign the
        values to A(1,1), A(2,1), A(3,1), etc.  Reading continues until n1 (10)
        successive row elements are filled.  Once the maximum row number is
        reached, subsequent data will be read into the next column (e.g.,
        A(1,2), A(2,2), A(3,2), etc.)

        For numerical parameters, absolute values and scale factors may be
        applied to the result parameter [``*VABS``, ``*VFACT``].  Results may be
        cumulative [``*VCUM``].  See the ``*VOPER`` command for details.  If you are in
        the GUI the ``*VREAD`` command must be contained in an externally prepared
        file read into the ANSYS program (i.e., ``*USE``, /INPUT, etc.).

        This command is not applicable to 4- or 5-D arrays.

        This command is valid in any processor.
        """
        command = f"*VREAD,{parr},{fname},{ext},,{label},{n1},{n2},{n3},{nskip}"
        return self.run(command, **kwargs)



# Copyright (C) 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os

import click

_USING_PIPE = [False]


def get_input_source(ctx, param, value):
    if not value and not click.get_text_stream("stdin").isatty():
        _USING_PIPE[0] = True
        return click.get_text_stream("stdin").read().strip()
    else:
        return value


@click.command(
    short_help="Convert APDL code to PyMAPDL code.",
    help="""PyMAPDL CLI tool for converting MAPDL scripts to PyMAPDL scripts.

    USAGE:

    This example demonstrates the main use of this tool:

        $ pymapdl convert mapdl.dat -o python.py

        File mapdl.dat successfully converted to python.py.

    The output argument is optional, in which case the "py" extension is used:

        $ pymapdl convert mapdl.dat

        File mapdl.dat successfully converted to mapdl.py.

    You can use any option from ``ansys.mapdl.core.convert.convert_script`` function:

        $ pymapdl convert mapdl.dat --auto-exit False

        File mapdl.dat successfully converted to mapdl.py.

        $ pymapdl convert mapdl.dat --filename_out mapdl.out --add_imports False

        File mapdl.dat successfully converted to mapdl.out.""",
)
@click.argument("filename_in", callback=get_input_source, required=False)
@click.option("-o", default=None, help="Name of the output Python script.")
@click.option("--filename_out", default=None, help="Name of the output Python script.")
@click.option(
    "--loglevel",
    default="WARNING",
    help="Logging level of the ansys object within the script.",
)
@click.option(
    "--auto_exit",
    default=True,
    type=bool,
    help="Adds a line to the end of the script to exit MAPDL. Default ``True``",
)
@click.option(
    "--line_ending", type=bool, default=None, help="When None, automatically is ``\n.``"
)
@click.option(
    "--exec_file",
    default=None,
    type=str,
    help="Specify the location of the ANSYS executable and include it in the converter output ``launch_mapdl`` call.",
)
@click.option(
    "--macros_as_functions",
    default=True,
    type=bool,
    help="Attempt to convert MAPDL macros to python functions.",
)
@click.option(
    "--use_function_names",
    default=True,
    type=bool,
    help="Convert MAPDL functions to ansys.mapdl.core.Mapdl class methods.  When ``True``, the MAPDL command ``K`` will be converted to ``mapdl.k``.  When ``False``, it will be converted to ``mapdl.run('k')``.",
)
@click.option(
    "--show_log",
    default=False,
    type=bool,
    help="Print the converted commands using a logger (from ``logging`` Python module).",
)
@click.option(
    "--add_imports",
    default=True,
    type=bool,
    help='If ``True``, add the lines ``from ansys.mapdl.core import launch_mapdl`` and ``mapdl = launch_mapdl(loglevel="WARNING")`` to the beginning of the output file. This option is useful if you are planning to use the output script from another mapdl session. See examples section. This option overrides ``auto_exit``.',
)
@click.option(
    "--comment_solve",
    default=False,
    type=bool,
    help='If ``True``, it will pythonically comment the lines that contain ``"SOLVE"`` or ``"/EOF"``.',
)
@click.option(
    "--cleanup_output",
    default=True,
    type=bool,
    help="If ``True`` the output is formatted using ``autopep8`` before writing the file or returning the string. This requires ``autopep8`` to be installed.",
)
@click.option(
    "--header",
    default=True,
    help="If ``True``, the default header is written in the first line of the output. If a string is provided, this string will be used as header.",
)
@click.option(
    "--print_com",
    default=True,
    type=bool,
    help="Print command ``/COM`` arguments to python console. Defaults to ``True``.",
)
@click.option(
    "--only_commands",
    default=False,
    type=bool,
    help="""If ``True``, it converts only the commands, meaning that header
(``header=False``), imports (``add_imports=False``),
and exit commands are NOT included (``auto_exit=False``).
Overrides ``header``, ``add_imports`` and ``auto_exit``.""",
)
@click.option(
    "--use_vtk",
    default=None,
    type=bool,
    help="""It sets the `mapdl.use_vtk` argument equals True or False depending on
this value. Defaults to `None` which is Mapdl class default.""",
)
@click.option(
    "--clear_at_start",
    default=False,
    help="""Add a `mapdl.clear()` after the Mapdl object initialization. Defaults to
`False`.""",
)
@click.option(
    "--check_parameter_names",
    default=False,
    help="""Set MAPDL object to avoid parameter name checks (do not raise leading underscored parameter exceptions). Defaults to `False`.""",
)
def convert(
    filename_in: str,
    o: str,
    filename_out: str,
    loglevel: str,
    auto_exit: bool,
    line_ending: str,
    exec_file: str,
    macros_as_functions: bool,
    use_function_names: bool,
    show_log: bool,
    add_imports: bool,
    comment_solve: bool,
    cleanup_output: bool,
    header: str,
    print_com: bool,
    only_commands: bool,
    use_vtk: bool,
    clear_at_start: bool,
    check_parameter_names: bool,
):
    """Convert MAPDL code to PyMAPDL"""
    from ansys.mapdl.core.convert import convert_apdl_block, convert_script

    if o:
        filename_out = o

    if _USING_PIPE[0]:
        code_block = filename_in  # from PIPE
        click.echo(
            convert_apdl_block(
                code_block,
                loglevel=loglevel,
                auto_exit=auto_exit,
                line_ending=line_ending,
                exec_file=exec_file,
                macros_as_functions=macros_as_functions,
                use_function_names=use_function_names,
                show_log=show_log,
                add_imports=add_imports,
                comment_solve=comment_solve,
                cleanup_output=cleanup_output,
                header=header,
                print_com=print_com,
                only_commands=only_commands,
                use_vtk=use_vtk,
                clear_at_start=clear_at_start,
                check_parameter_names=check_parameter_names,
            )
        )

    else:
        convert_script(
            filename_in,
            filename_out,
            loglevel,
            auto_exit,
            line_ending,
            exec_file,
            macros_as_functions,
            use_function_names,
            show_log,
            add_imports,
            comment_solve,
            cleanup_output,
            header,
            print_com,
            only_commands,
            use_vtk,
            clear_at_start,
            check_parameter_names,
        )

        if filename_out:
            click.echo(
                click.style("Success: ", fg="green")
                + f"File {filename_in} successfully converted to {filename_out}."
            )
        else:
            click.echo(
                click.style("Success: ", fg="green")
                + f"File {filename_in} successfully converted to {os.path.splitext(filename_in)[0] + '.py'}."
            )



# Copyright (C) 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from logging import Logger, StreamHandler
import os
import re
from warnings import warn

from ansys.mapdl.core import __version__
from ansys.mapdl.core.commands import Commands
from ansys.mapdl.core.misc import is_float

# Because the APDL version has empty arguments, whereas the PyMAPDL
# doesn't have them. Hence the order of arguments is messed up.

FORMAT_OPTIONS = {
    "select": "W191,W291,W293,W391,E115,E117,E122,E124,E125,E225,E231,E301,E303,F401,F403",
    "max-line-length": 100,
}

LOGLEVEL_DEFAULT = "WARNING"
AUTO_EXIT_DEFAULT = True
LINE_ENDING_DEFAULT = None
EXEC_FILE_DEFAULT = None
MACROS_AS_FUNCTIONS_DEFAULT = True
USE_FUNCTION_NAMES_DEFAULT = True
SHOW_LOG_DEFAULT = False
ADD_IMPORTS_DEFAULT = True
COMMENT_SOLVE_DEFAULT = False
CLEANUP_OUTPUT_DEFAULT = True
HEADER_DEFAULT = True
PRINT_COM_DEFAULT = True
ONLY_COMMANDS_DEFAULT = False
USE_VTK_DEFAULT = None
CLEAR_AT_START_DEFAULT = False
CHECK_PARAMETER_NAMES_DEFAULT = True


# This commands have "--" as one or some arguments
COMMANDS_WITH_EMPTY_ARGS = {
    "/CMA": (),  # "/CMAP,
    "/NER": (),  # "/NERR,
    "/PBF": (),  # "/PBF,
    "/PMO": (),  # "/PMORE,
    "ADD": (),  # "ADD"
    "ANTY": (),  # ANTYPE,
    "ASBL": (),  # ASBL,
    "ATAN": (),  # ATAN,
    "BCSO": (),  # BCSOPTION,
    "CORI": (),  # CORIOLIS
    "CDRE": (),  # CDREAD
    "CLOG": (),  # CLOG,
    "CONJ": (),  # CONJUG,
    "DERI": (),  # DERIV,
    "DSPO": (),  # DSPOPTION,
    "ENER": (),  # ENERSOL,
    "ENSY": (),  # ENSYM,
    "ESYM": (),  # ESYM,
    "EXP": (),  # EXP,
    "EXPA": (),  # EXPAND,
    "FCLI": (),  # FCLIST,
    "FILE": (),  # FILEAUX2,
    "FLUR": (),  # FLUREAD,
    "GMAT": (),  # GMATRIX,
    "IMAG": (),  # IMAGIN,
    "INT1": (),  # INT1,
    "LARG": (),  # LARGE,
    "LATT": (),  # LATT,
    "MAP": (),  # MAP,
    "MORP": (),  # MORPH,
    "MPCO": (),  # MPCOPY,
    "NLOG": (),  # NLOG,
    "PLMA": (),  # PLMAP,
    "PRED": (),  # PRED,
    "PROD": (),  # PROD,
    "QRDO": (),  # QRDOPT,
    "QUOT": (),  # QUOT,
    "RACE": (),  # RACE,
    "RDEC": (),  # RDEC
    "REAL": (),  # REALVAR,
    "REME": (),  # REMESH,
    "RPSD": (),  # RPSD
    "SECR": (),  # SECREAD
    "SECW": (),  # SECWRITE
    "SESY": (),  # SESYMM,
    "SETF": (),  # SETFGAP,
    "SETR": (),  # SETRAN,
    "SMAL": (),  # SMALL,
    "SNOP": (),  # SNOPTION,
    "SQRT": (),  # SQRT
    "SURE": (),  # SURESU,
    "THOP": (),  # THOPT,
    "TINT": (),  # TINTP,
    "XFDA": (),  # XFDATA
}


COMMANDS_TO_NOT_BE_CONVERTED = [
    "CMPL",  # CMPLOT default behaviour does not match the `mapdl.cmplot`'s at the moemnt
    # CDREAD # commented above
]

FORCED_MAPPING = {
    # Forced mapping between MAPDL and PyMAPDL
    "SECT": "sectype",  # Because it is shadowed by `sectinqr`
}


def convert_script(
    filename_in,
    filename_out=None,
    loglevel="WARNING",
    auto_exit=True,
    line_ending=None,
    exec_file=None,
    macros_as_functions=True,
    use_function_names=True,
    show_log=False,
    add_imports=True,
    comment_solve=False,
    cleanup_output=True,
    header=True,
    print_com=True,
    only_commands=False,
    use_vtk=None,
    clear_at_start=False,
    check_parameter_names=True,
):
    """Converts an ANSYS input file to a python PyMAPDL script.

    Parameters
    ----------
    filename_in : str
        Filename of the ansys input file to read in.

    filename_out : str, optional
        Filename of the python script to write a translation to.
        Defaults to the ``filename_in`` name ending in ``py``.

    loglevel : str, optional
        Logging level of the ansys object within the script.

    auto_exit : bool, optional
        Adds a line to the end of the script to exit MAPDL.  Default
        ``True``.

    line_ending : str, optional
        When None, automatically is ``\n.``

    exec_file : str, optional
        Specify the location of the ANSYS executable and include
        it in the converter output ``launch_mapdl`` call.

    macros_as_functions : bool, optional
        Attempt to convert MAPDL macros to python functions.

    use_function_names : bool, optional
        Convert MAPDL functions to ansys.mapdl.core.Mapdl class
        methods.  When ``True``, the MAPDL command "K" will be
        converted to ``mapdl.k``.  When ``False``, it will be
        converted to ``mapdl.run('k')``.

    show_log : bool, optional
        Print the converted commands using a logger (from ``logging``
        Python module).

    add_imports : bool, optional
        If ``True``, add the lines ``from ansys.mapdl.core import launch_mapdl``
        and ``mapdl = launch_mapdl(loglevel="WARNING")`` to the
        beginning of the output file. This option is useful if you
        are planning to use the output script from another mapdl
        session. See examples section.
        This option overrides ``auto_exit``.

    comment_solve : bool, optional
        If ``True``, it will pythonically comment the lines that
        contain ``"SOLVE"`` or ``"/EOF"``.

    cleanup_output : bool, optional
        If ``True`` the output is formatted using ``autopep8`` before writing
        the file or returning the string. This requires ``autopep8`` to be
        installed.

    header : bool, optional
        If ``True``, the default header is written in the first line
        of the output. If a string is provided, this string will be
        used as header.

    print_com : bool, optional
        Print command ``/COM`` arguments to python console.
        Defaults to ``True``.

    only_commands : bool, optional
        If ``True``, it converts only the commands, meaning that header
        (``header=False``), imports (``add_imports=False``),
        and exit commands are NOT included (``auto_exit=False``).
        Overrides ``header``, ``add_imports`` and ``auto_exit``.

    use_vtk : bool, optional
        It sets the `mapdl.use_vtk` argument equals True or False depending on
        this value.

    clear_at_start : bool, optional
        Add a `mapdl.clear()` after the Mapdl object initialization.

    Returns
    -------
    list
        List of lines translated.

    Examples
    --------
    >>> from ansys.mapdl import core as pymapdl
    >>> from ansys.mapdl.core import examples
    >>> clines = pymapdl.convert_script(examples.vmfiles['vm1'], 'vm1.py')

    Converting a script and using it already in the same session.
    For this case, it is recommended to use
    :func:`convert_apdl_block() <ansys.mapdl.core.convert_apdl_block>`
    since you do not need to write the file.

    >>> import os
    >>> from ansys.mapdl.core import launch_mapdl
    >>> from ansys.mapdl.core import examples
    >>> from ansys.mapdl.core import convert_script
    >>> in_file = examples.vmfiles['vm10']
    >>> filename = os.path.basename(in_file)
    >>> out_file = 'out_' + filename.replace('.dat', '.py')
    >>> output = convert_script(file, out_file, line_ending='\\n')
    >>> mapdl = launch_mapdl()
    >>> with open(out_file, 'r') as fid:
    ...     cmds = fid.read()
    >>> mapdl.input_strings(cmds.splitlines()[2:10])

    """
    with open(filename_in, "r") as fid:
        apdl_strings = fid.readlines()

    if filename_out is None:
        filename_out = os.path.splitext(filename_in)[0] + ".py"
        if os.path.exists(filename_out):
            raise FileExistsError(
                f"The predefined output file '{filename_out}' already exists."
            )

    translator = _convert(
        apdl_strings=apdl_strings,
        loglevel=loglevel,
        auto_exit=auto_exit,
        line_ending=line_ending,
        exec_file=exec_file,
        macros_as_functions=macros_as_functions,
        use_function_names=use_function_names,
        show_log=show_log,
        add_imports=add_imports,
        comment_solve=comment_solve,
        cleanup_output=cleanup_output,
        header=header,
        print_com=print_com,
        only_commands=only_commands,
        use_vtk=use_vtk,
        clear_at_start=clear_at_start,
        check_parameter_names=check_parameter_names,
    )

    translator.save(filename_out)
    return translator.lines


def convert_apdl_block(
    apdl_strings,
    loglevel="WARNING",
    auto_exit=True,
    line_ending=None,
    exec_file=None,
    macros_as_functions=True,
    use_function_names=True,
    show_log=False,
    add_imports=True,
    comment_solve=False,
    cleanup_output=True,
    header=True,
    print_com=True,
    only_commands=False,
    use_vtk=None,
    clear_at_start=False,
    check_parameter_names=False,
):
    """Converts an ANSYS input string to a python PyMAPDL string.

    Parameters
    ----------
    apdl_string : str
        APDL strings or list of strings to convert.

    loglevel : str, optional
        Logging level of the ansys object within the script.

    auto_exit : bool, optional
        Adds a line to the end of the script to exit MAPDL.  Default
        ``True``.

    line_ending : str, optional
        When None, automatically determined by OS being used.

    macros_as_functions : bool, optional
        Attempt to convert MAPDL macros to python functions.

    use_function_names : bool, optional
        Convert MAPDL functions to ansys.mapdl.core.Mapdl class
        methods.  When ``True``, the MAPDL command "K" will be
        converted to ``mapdl.k``.  When ``False``, it will be
        converted to ``mapdl.run('k')``.

    show_log : bool, optional
        Print the converted commands using a logger (from ``logging``
        Python module).

    add_imports : bool, optional
        If ``True``, it add the next lines to the beginning of the
        output file:

        .. code:: python

           from ansys.mapdl.core import launch_mapdl

           mapdl = launch_mapdl(loglevel="WARNING")

        This option is useful if you are planning to use the output
        script from another mapdl session. See examples section.
        This option overrides ``'auto_exit'``.

    comment_solve : bool, optional
        If ``True``, pythonically comment the lines containing
        ``mapdl.solve`` or ``"/EOF"``.

    cleanup_output : bool, optional
        If ``True`` the output is formatted using ``autopep8`` before
        writing the file or returning the string.

    header : bool, optional
        If ``True``, the default header is written in the first line
        of the output. If a string is provided, this string will be
        used as header.
    print_com : bool, optional
        Print command ``/COM`` arguments to python console.
        Defaults to ``True``.

    only_commands : bool, optional
        If ``True``, it converts only the commands, meaning that header
        (``header=False``), imports (``add_imports=False``),
        and exit commands are NOT included (``auto_exit=False``).
        Overrides ``header``, ``add_imports`` and ``auto_exit``.

    use_vtk : bool, optional
        It sets the `mapdl.use_vtk` argument equals True or False depending on
        this value. Defaults to `None` which is Mapdl class default.

    clear_at_start : bool, optional
        Add a `mapdl.clear()` after the Mapdl object initialization. Defaults to
        `False`.

    check_parameter_names : bool, optional
        Set MAPDL object to avoid parameter name checks (do not raise leading underscored parameter exceptions). Defaults to `False`.

    Returns
    -------
    list
        List of lines translated.

    Examples
    --------
    Convert a script and use it in the same session.

    >>> from ansys.mapdl.core import examples, launch_mapdl, convert_apdl_block
    >>> in_file = examples.vmfiles['vm10']
    >>> filename = in_file.split('\\')[-1]
    >>> out_file = 'out_' + filename.replace('.dat', '.py')
    >>> cmds = convert_apdl_block(file, out_file, line_ending='\\n')
    >>> # Do any change in the text, for example:
    >>> cmds = cmds.replace('solve', '!solve')
    >>> mapdl = launch_mapdl()
    >>> mapdl.input_strings(cmds.splitlines()[2:10])

    """

    translator = _convert(
        apdl_strings,
        loglevel=loglevel,
        auto_exit=auto_exit,
        line_ending=line_ending,
        exec_file=exec_file,
        macros_as_functions=macros_as_functions,
        use_function_names=use_function_names,
        show_log=show_log,
        add_imports=add_imports,
        comment_solve=comment_solve,
        cleanup_output=cleanup_output,
        header=header,
        print_com=print_com,
        only_commands=only_commands,
        use_vtk=use_vtk,
        clear_at_start=clear_at_start,
        check_parameter_names=check_parameter_names,
    )

    if isinstance(apdl_strings, str):
        return translator.line_ending.join(translator.lines)
    return translator.lines


def _convert(
    apdl_strings,
    loglevel="WARNING",
    auto_exit=True,
    line_ending=None,
    exec_file=None,
    macros_as_functions=True,
    use_function_names=True,
    show_log=False,
    add_imports=True,
    comment_solve=False,
    cleanup_output=True,
    header=True,
    print_com=True,
    only_commands=False,
    use_vtk=None,
    clear_at_start=False,
    check_parameter_names=True,
):
    if only_commands:
        auto_exit = False
        add_imports = False
        header = False

    translator = FileTranslator(
        loglevel,
        line_ending,
        exec_file=exec_file,
        macros_as_functions=macros_as_functions,
        use_function_names=use_function_names,
        show_log=show_log,
        add_imports=add_imports,
        comment_solve=comment_solve,
        cleanup_output=cleanup_output,
        header=header,
        print_com=print_com,
        use_vtk=use_vtk,
        clear_at_start=clear_at_start,
        check_parameter_names=check_parameter_names,
    )

    if isinstance(apdl_strings, str):
        # os.linesep does not work well, so we are making sure
        # the line separation is appropriate.
        regx = f"[^\\r]({translator.line_ending})"
        if not re.search(regx, apdl_strings):
            if "\r\n" in apdl_strings:
                translator.line_ending = "\r\n"
            elif "\n" in apdl_strings:
                translator.line_ending = "\n"

        apdl_strings = apdl_strings.split(translator.line_ending)

    for line in apdl_strings:
        translator.translate_line(line)

    if auto_exit and add_imports:
        translator.write_exit()
    return translator


class Lines(list):
    def __init__(self, mute):
        self._log = Logger("convert_logger")
        self._setup_logger()
        self._mute = mute
        super().__init__()

    def append(self, item, mute=True):
        # append the item to itself (the list)
        if not self._mute and item:
            stripped_msg = item.replace("\n", "\\n")
            self._log.info(msg=f"Converted: '{stripped_msg}'")
        super(Lines, self).append(item)

    def _setup_logger(self):
        stdhdl = StreamHandler()
        stdhdl.setLevel(10)
        stdhdl.set_name("stdout")
        self._log.addHandler(stdhdl)
        self._log.propagate = True


class FileTranslator:
    obj_name = "mapdl"
    indent = ""
    non_interactive = False

    def __init__(
        self,
        loglevel="INFO",
        line_ending=None,
        exec_file=None,
        macros_as_functions=True,
        use_function_names=True,
        show_log=False,
        add_imports=True,
        comment_solve=False,
        cleanup_output=True,
        header=True,
        print_com=True,
        use_vtk=None,
        clear_at_start=False,
        check_parameter_names=False,
    ):
        self._non_interactive_level = 0
        self.lines = Lines(mute=not show_log)
        self._functions = []
        if line_ending:
            self.line_ending = line_ending
        else:
            self.line_ending = "\n"
        self.macros_as_functions = macros_as_functions
        self._infunction = False
        self.use_function_names = use_function_names
        self.comment = ""
        self._add_imports = add_imports
        self._comment_solve = comment_solve
        self.cleanup_output = cleanup_output
        self._header = header
        self.print_com = print_com
        self.verification_example = False
        self.use_vtk = use_vtk
        self.clear_at_start = clear_at_start
        self.check_parameter_names = check_parameter_names
        self.macros_names = []

        self.write_header()
        if self._add_imports:
            self.initialize_mapdl_object(loglevel, exec_file)

        self._valid_commands = self._get_valid_pymapdl_methods_short()
        self._block_commands = {
            "NBLO": "NBLOCK",
            "EBLO": "EBLOCK",
            "BFBL": "BFBLOCK",
            "BFEB": "BFEBLOCK",
            "PREA": "PREAD",
            "SFEB": "SFEBLOCK",
        }  # Way out: '-1' , 'END PREAD'

        self._enum_block_commands = {
            "CMBL": "CMBLOCK",
        }  # Commands where you need to count the number of lines.

        _NON_INTERACTIVE_COMMANDS = {
            "*VWR": "*VWRITE",
            "*VRE": "*VREAD",
        }

        self._non_interactive_commands = (
            list(_NON_INTERACTIVE_COMMANDS)
            + list(self._block_commands)
            + list(self._enum_block_commands)
        )

        self._chained_commands = 0
        self.chained_commands = False
        self._block_count = 0
        self._block_count_target = 0
        self._in_block = False
        self._block_current_cmd = None

    def write_header(self):
        if isinstance(self._header, bool):
            if self._header:
                header = f'"""Script generated by ansys-mapdl-core version {__version__}"""\n'
                self.lines.append(header)
        elif isinstance(self._header, str):
            self.lines.append(f'"""{self._header}"""\n')

        else:
            raise TypeError(
                "The keyword argument 'header' should be a string or a boolean."
            )

    def write_exit(self):
        self.lines.append(f"\n{self.obj_name}.exit()")

    def format_using_autopep8(self, text=None):
        """Format internal `self.lines` with autopep8.

        Parameters
        ----------
        text : str, optional
            Text to format instead of `self.lines`. For development use.

        """
        if self.cleanup_output:
            try:
                import autopep8
            except ModuleNotFoundError:  # pragma: no cover
                warn(
                    "Install `autopep8` to use this feature with\n"
                    "`pip install autopep8`"
                )
                return

            if not text:
                text = self.line_ending.join(self.lines)
                self.lines = autopep8.fix_code(text).splitlines()

            else:  # pragma: no cover
                # for development purposes
                return autopep8.fix_code(text)

    def save(self, filename, format_autopep8=True):
        """Saves lines to file"""
        if os.path.isfile(filename):
            os.remove(filename)

        # Making sure we write python string with double slash.
        # We are not expecting other type of unicode symbols.
        self.lines = [each_line.replace("\\", "\\\\") for each_line in self.lines]

        # Try to format the file using AutoPEP8
        self.format_using_autopep8()

        with open(filename, "w") as f:
            f.write(self.line_ending.join(self.lines))

    def initialize_mapdl_object(self, loglevel, exec_file):
        """Initializes ansys object as lines"""
        core_module = "ansys.mapdl.core"  # shouldn't change
        self.lines.append(f"from {core_module} import launch_mapdl")

        mapdl_arguments = [f'loglevel="{loglevel}"']

        if exec_file:
            mapdl_arguments.append(f'exec_file="{exec_file}"')

        if self.print_com:
            mapdl_arguments.append("print_com=True")

        if self.use_vtk is not None:
            mapdl_arguments.append(f"use_vtk={bool(self.use_vtk)}")

        if self.check_parameter_names is not None and not self.check_parameter_names:
            mapdl_arguments.append("check_parameter_names=False")

        line = f'{self.obj_name} = launch_mapdl({", ".join(mapdl_arguments)})'
        self.lines.append(line)

        if self.clear_at_start:
            self.lines.append(f"{self.obj_name}.clear() # Clearing session")

    @property
    def line_ending(self):
        return self._line_ending

    @line_ending.setter
    def line_ending(self, line_ending):
        if line_ending not in ["\n", "\r\n"]:
            raise ValueError('Line ending must be either "\\n", "\\r\\n"')
        self._line_ending = line_ending

    def translate_line(self, line):
        """Converts a single line from an ANSYS APDL script"""

        if "$" in line:
            # these are chained commands.
            lines = line.split("$")
            self.start_chained_commands()
            for each_line in lines:
                self.translate_line(each_line)

            self.end_chained_commands()
            return

        self.comment = ""
        original_line = line.replace("\r\n", "").replace(
            "\n", ""
        )  # It is needed for the nblock, eblock since they have spaces before the numbers
        line = line.strip()
        line = line.replace('"', "'")

        if self._in_block:
            self._block_count += 1

        if (
            self._in_block
            and self._block_count >= self._block_count_target
            and self._block_count_target
        ):
            self._in_block = False
            self.end_non_interactive()
            self._block_count = 0
            self._block_current_cmd = None

        # check if line contains a comment
        if "!" in line:
            if "'!'" in line or '"!"' in line:
                pass
            elif line[0] == "!":  # entire line is a comment
                self.comment = line.replace("!", "").strip()
                self.store_comment()
                return
            else:  # command and in-line comment
                split_line = line.split("!")
                line = split_line[0]
                self.comment = " ".join(split_line[1:])
                self.comment = self.comment.lstrip()

        if not line:
            # Keeping empty lines
            self.lines.append("")
            return

        # Cleaning ending empty arguments.
        # Because of an extra comma added to toffst command when generating ds.dat.
        line_ = line.split(",")[::-1]  # inverting order

        for ind, each in enumerate(line_):
            if each.strip():  # strip to remove spaces in empty arguments
                break

        line = ",".join(line_[ind:][::-1])

        line_with_trailing_commas = line

        # remove trailing comma
        line = line[:-1] if line[-1] == "," else line
        line_upper = line.upper()

        cmd_caps = line.split(",")[0].upper()
        cmd_caps_short = cmd_caps[:4]

        items = self._get_items(line.strip())

        if cmd_caps_short in ["SOLV", "LSSO"] and self._comment_solve:
            self.store_command(
                "com",
                ["The following line has been commented due to `comment_solve`:"],
            )
            self.store_command("com", [line])
            return

        if cmd_caps_short == "/VER":
            self.verification_example = True

        if cmd_caps_short == "/COM":
            # It is a comment
            self.store_command("com", [line[5:]])
            return

        if cmd_caps_short == "C***":
            # It is an old style comment
            self.store_command("com", [line[5:]])
            return

        if cmd_caps == "*DO":
            self.start_non_interactive()
            self.store_run_command(line)
            return

        if cmd_caps in ["*ENDDO", "*ENDIF"]:
            self.store_run_command(line)
            self.end_non_interactive()
            return

        if "/EXI" in cmd_caps.upper() and self.non_interactive:
            self.store_command("com", [f"Skipping: {line}"])
            return

        if self.output_to_file(line):
            if self.verification_example and "SCRATCH" in line.upper():
                self.store_command("com", [f"Skipping: {line}"])
            else:
                self.start_non_interactive()
                self.store_run_command(line)
            return

        if self.output_to_default(line):
            self.store_run_command(line)
            self.store_run_command("/GOPR")  # Adding gopr to ensure printing out
            self.end_non_interactive()
            return

        if cmd_caps == "/VERIFY":
            self.store_run_command("FINISH")
            self.store_run_command(line)
            return

        if cmd_caps_short == "*REP":
            if not self.non_interactive:
                prev_cmd = self.lines.pop(-1)
                self.start_non_interactive()
                new_prev_cmd = (
                    self.indent + prev_cmd
                )  # Since we are writing in self.lines we need to add the indentation by ourselves.
                self.lines.append(new_prev_cmd)
                self.store_run_command(
                    line
                )  # Using run but it could be `store_command`
                self.end_non_interactive()
                return

        # Skipping conversion if command has empty arguments and there is ",," in the call
        if (
            cmd_caps_short in COMMANDS_WITH_EMPTY_ARGS
            and ",," in line_with_trailing_commas.replace(" ", "")
        ):
            self.store_run_command(line.strip())
            return

        # Skipping commands to not be converted
        if cmd_caps_short in COMMANDS_TO_NOT_BE_CONVERTED:
            self.store_run_command(line.strip())
            return

        if cmd_caps_short == "/TIT":  # /TITLE
            parameters = line.split(",")[1:]
            return self.store_command("title", [",".join(parameters).strip()])

        if cmd_caps_short == "/AXL":  # /AXLAB
            parameters = line.split(",")[1:]
            parameters_ = [parameters[0], ",".join(parameters[1:])]
            return self.store_command("axlab", parameters_)

        if cmd_caps_short == "*GET":
            if self.non_interactive:  # gives error
                self.store_run_command(line)
                return
            else:
                parameters = line.split(",")[1:]
                return self.store_command("get", parameters)

        if cmd_caps_short == "/NOP":
            self.comment = (
                "It is not recommended to use '/NOPR' in a normal PyMAPDL session."
            )
            self.store_under_scored_run_command(line)
            return

        if cmd_caps_short == "*CRE":  # creating a function
            self.macros_names.append(items[1])
            if self.macros_as_functions:
                self.start_function(items[1].strip())
                return
            else:
                self.start_non_interactive()

        ## Treating functions
        if items[0] in self.macros_names and self.macros_as_functions:
            # We are calling the function/macro created before.
            func_name = items[0].strip()

            args = self._parse_arguments(items[1:])
            self.store_python_command(f"{func_name}({args})")
            return

        if cmd_caps_short == "*USE" and self.macros_as_functions:
            func_name = items[1].strip()
            if func_name in self._functions:
                args = self._parse_arguments(items[2:])
                self.store_python_command(f"{func_name}({args})")
                return

        if cmd_caps == "/PREP7":
            return self.store_command("prep7", [])

        if "*END" in line_upper:
            if self.macros_as_functions:
                self.store_empty_line()
                self.store_empty_line()
                self.indent = self.indent[4:]
                self._infunction = False
                if not self._in_block:
                    self.end_non_interactive()
                return
            else:
                self.store_run_command(line)
                if not self._in_block:
                    self.end_non_interactive()
                return

        # check for if statement
        if cmd_caps[:3] == "*IF" or "*IF" in line_upper:
            self.start_non_interactive()
            self.store_run_command(line)
            return

        # check if line ends non-interactive
        if line[0] == "(":
            if not self.non_interactive:
                warn(
                    "\nPossible invalid line:\n%s\n" % line
                    + "This line requires a *VWRITE beforehand."
                    + "The previous line is: \n%s\n\n" % self.lines[-1]
                )
            self.store_run_command(line)
            return

        # check if a line is setting a variable
        if "=" in items[0]:  # line sets a variable:
            self.store_run_command(line)
            return

        command = items[0].lower().strip()
        parameters = items[1:]
        if not command:
            self.store_empty_line()
            return

        if line == "-1" or line_upper == "END PREAD":  # End of block commands
            self.store_run_command(line)
            self._in_block = False
            self.end_non_interactive()
            return

        # check valid command
        if (
            self._pymapdl_command(command) not in self._valid_commands
            and cmd_caps_short in self._non_interactive_commands
        ):
            if cmd_caps_short in self._block_commands:
                self._in_block = True
                self._block_count = 0
                self._block_count_target = 0

            elif cmd_caps_short in self._enum_block_commands:
                self._in_block = True
                self._block_count = 0
                if cmd_caps_short == "CMBL":  # In cmblock
                    # CMBLOCK,Cname,Entity,NUMITEMS,,,,,KOPT
                    numitems = int(line.split(",")[3])
                    _block_count_target = (
                        numitems // 8 + 1 if numitems % 8 != 0 else numitems // 8
                    )
                    self._block_count_target = (
                        _block_count_target + 2
                    )  # because the cmd_caps_short line and option line.

            self._block_current_cmd = cmd_caps_short
            self.start_non_interactive()
            self.store_run_command(line.strip())

        elif self.use_function_names:
            # Takign into account the leading characters
            if command.upper() in FORCED_MAPPING:
                # Checking exceptions/forced mapping
                command = FORCED_MAPPING[command.upper()]

            else:
                # Looking for a suitable candidate.
                if command[0] == "/":
                    slash_command = f"slash{command[1:4]}"
                    if slash_command in self._valid_commands:
                        command = slash_command
                    else:
                        command = command[1:]
                elif command[0] == "*":
                    star_command = f"star{command[1:4]}"
                    if star_command in self._valid_commands:
                        command = star_command
                    else:
                        command = command[1:]

                # Some commands are abbreviated (only 4 letters)
                from ansys.mapdl.core import Mapdl

                if command not in dir(Mapdl):
                    command = self.find_match(command)

            # Storing
            if command:
                self.store_command(command, parameters)
            else:  # find_match can return None
                self.store_run_command(line.strip())

        else:
            self.store_run_command(line.strip())

    def _pymapdl_command(self, command):
        if command[0] in ["/", "*"]:
            command = command[1:]
        return command

    def start_function(self, func_name):
        self._functions.append(func_name)
        self.store_empty_line()
        self.store_empty_line()
        self._infunction = True
        spacing = " " * (len(func_name) + 5)
        line = self.indent + "def %s(%s," % (
            func_name,
            ", ".join(["ARG%d=''" % i for i in range(1, 7)]),
        )
        line += "\n"
        line += "%s%s," % (
            spacing,
            ", ".join(["ARG%d=''" % i for i in range(7, 13)]),
        )
        line += "\n"
        line += "%s%s):" % (
            spacing,
            ", ".join(["ARG%d=''" % i for i in range(13, 19)]),
        )
        self.lines.append(line)
        self.indent = self.indent + "    "

    def store_under_scored_run_command(self, command):
        self.store_run_command(command, run_underscored=True)

    def store_run_command(self, command, run_underscored=False):
        """Stores pyansys.ANSYS command that cannot be broken down
        into a function and parameters.
        """
        if run_underscored:
            underscore = "_"
        else:
            underscore = ""

        # Removing trailing/leading spaces
        command = command.strip()

        if self._infunction and "ARG" in command:
            args = []
            for i in range(1, 19):
                arg = "ARG%d" % i
                c = 0
                if arg in command:
                    command = command.replace(arg, "{" + f"{arg}" + "}")
                    args.append(arg)
                    c += 1

            line = '%s%s.%srun(f"%s")' % (
                self.indent,
                self.obj_name,
                underscore,
                command,
            )

        elif self.comment:
            line = '%s%s.%srun("%s")  # %s' % (
                self.indent,
                self.obj_name,
                underscore,
                command,
                self.comment,
            )
        else:
            line = '%s%s.%srun("%s")' % (
                self.indent,
                self.obj_name,
                underscore,
                command,
            )
        self.lines.append(line)

    def store_comment(self):
        """Stores a line containing only a comment"""
        line = f"{self.indent}# {self.comment}"
        self.lines.append(line)

    def store_empty_line(self):
        """Stores an empty line"""
        self.lines.append("")

    def store_python_command(self, command):
        line = f"{self.indent}{command}"
        self.lines.append(line)

    def _parse_arguments(self, parameters):
        parsed_parameters = []
        for parameter in parameters:
            parameter = parameter.strip()
            if is_float(parameter) and parameter.upper() != "INF":
                parsed_parameters.append(parameter)
            elif "ARG" in parameter and self._infunction:
                parsed_parameters.append("%s" % parameter)
            else:
                # Removing strings '' and "" because they are going to be added by the converter module.
                if parameter.startswith("'") and parameter.endswith("'"):
                    parameter = parameter[1:-1]
                if parameter.startswith('"') and parameter.endswith('"'):
                    parameter = parameter[1:-1]
                parsed_parameters.append(f'"{parameter}"')

        return ", ".join(parsed_parameters)

    def store_command(self, function, parameters):
        """Stores a valid pyansys function with parameters"""
        parameter_str = self._parse_arguments(parameters)

        if self.comment:
            line = "%s%s.%s(%s)  # %s" % (
                self.indent,
                self.obj_name,
                function,
                parameter_str,
                self.comment,
            )
        else:
            line = "%s%s.%s(%s)" % (
                self.indent,
                self.obj_name,
                function,
                parameter_str,
            )

        self.lines.append(line)

    def start_non_interactive(self):
        self._non_interactive_level += 1
        if self.non_interactive:
            return
        line = f"{self.indent}with {self.obj_name}.non_interactive:"
        self.lines.append(line)
        self.non_interactive = True
        self.indent = self.indent + "    "

    def end_non_interactive(self):
        self._non_interactive_level -= 1
        if self._non_interactive_level <= 0:
            self.indent = self.indent[4:]
            self.non_interactive = False

    def start_chained_commands(self):
        self._chained_commands += 1
        if self.chained_commands:
            return
        line = f"{self.indent}with {self.obj_name}.chain_commands:"
        self.lines.append(line)
        self.chained_commands = True
        self.indent = self.indent + "    "

    def end_chained_commands(self):
        self._chained_commands -= 1
        if self._chained_commands <= 0:
            self.indent = self.indent[4:]
            self.chained_commands = False

    def output_to_file(self, line):
        """Return if an APDL line is redirecting to a file."""
        if line[:4].upper() == "/OUT":
            # We are redirecting the output to somewhere, probably a file.
            # Because of the problem with the ansys output, we need to execute
            # this in non_interactive mode.
            output_cmd = line.strip().upper().split(",")
            if len(output_cmd) > 1:
                opt1 = output_cmd[1].strip().upper()
                if opt1 != "TERM":
                    # A file is supplied.
                    return True

        if (
            line[:4].upper() == "*CFO"
        ):  # any([each[0:4] in '*CFOPEN' for each in dir(Commands)])
            # We might not need going into interactive mode for *CFOPEN/*CFCLOSE
            return True

        return False

    def output_to_default(self, line):
        if line[:4].upper() == "/OUT":
            # We are redirecting the output to somewhere, probably a file.
            # Because of the problem with the ansys output, we need to execute
            # this in non_interactive mode.
            output_cmd = line.strip().upper().split(",")
            if len(output_cmd) == 1:
                return True
            elif len(output_cmd) > 1:
                opt1 = output_cmd[1].strip().upper()
                if opt1 == "TERM":
                    # A file is supplied.
                    return True
        if line[:4].upper() in "*CFCLOSE":
            # We might not need going into interactive mode for *CFOPEN/*CFCLOSE
            return True

        return False

    def _get_items(self, line_):
        """Parse the line items (comma separated elements) but ignoring the ones inside parenthesis, or brackets"""

        parenthesis_count = 0

        items = []
        begining_substring = 0

        for ind, each_char in enumerate(line_):
            if each_char in ["(", "[", "{"]:
                parenthesis_count += 1

            if each_char == "," and parenthesis_count == 0:
                items.append(line_[begining_substring:ind])
                begining_substring = ind + 1

            if ind == len(line_) - 1:  # reaching ending of line
                items.append(line_[begining_substring : ind + 1])

            if each_char in [")", "]", "}"]:
                parenthesis_count -= 1

        return items

    def _get_valid_pymapdl_methods_short(self):
        pymethods = dir(Commands)

        reduced_list = []
        for each_method in pymethods:
            if not re.match(r"^[\*~/A-Za-z]\w*$", each_method):
                continue
            if each_method.startswith("slash"):
                reduced_list.append(each_method[:8])
            elif each_method.startswith("star"):
                reduced_list.append(each_method[:7])
            else:
                reduced_list.append(each_method[:4])
        return reduced_list

    def find_match(self, cmd):
        pymethods = sorted(dir(Commands))

        for each in pymethods:
            if each.startswith(cmd):
                return each



# Copyright (C) 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Store parameters for a PyMAPDL-specific theme for pyvista"""

import numpy as np

try:
    from matplotlib.colors import ListedColormap

    _HAS_MATPLOTLIB = True
except ModuleNotFoundError:
    _HAS_MATPLOTLIB = False


from ansys.mapdl.core import _HAS_PYVISTA

if _HAS_PYVISTA:
    from pyvista.plotting.colors import get_cycler

    try:
        from pyvista.plotting.themes import Theme

    except ImportError:
        from pyvista import __version__ as pyvista_version

        if "0.40" in pyvista_version:
            from pyvista.themes import Theme
        else:  # older versions
            from pyvista.themes import DefaultTheme as Theme

else:  # pragma: no cover

    class Theme:
        pass


MAPDL_colorbar = (
    np.array(
        [
            [0, 0, 255],
            [0, 178, 255],
            [0, 255, 255],
            [0, 255, 178],
            [0, 255, 0],
            [178, 255, 0],
            [255, 255, 0],
            [255, 178, 0],
            [255, 0, 0],
        ],
        dtype=float,
    )
    / 255
)

if _HAS_MATPLOTLIB:
    PyMAPDL_cmap: ListedColormap = ListedColormap(MAPDL_colorbar, name="PyMAPDL")


def get_ansys_colors(N=9):
    if not _HAS_MATPLOTLIB:
        raise ModuleNotFoundError(
            "'matplotlib' package is needed for 'get_ansys_colors'."
        )
    return np.array([PyMAPDL_cmap(i) for i in range(N)])


class MapdlTheme(Theme):
    """PyMAPDL-specific theme for Pyvista.

    Examples
    --------
    Create a custom theme with unique parameters from the base MapdlTheme.

    >>> from ansys.mapdl import core as pymapdl
    >>> my_theme = pymapdl.MapdlTheme()
    >>> my_theme.background = 'white'
    >>> my_theme.cmap = 'jet'  # colormap
    >>> my_theme.axes.show = False
    >>> my_theme.show_scalar_bar = False

    Apply this theme to element plotting.

    >>> mapdl.eplot(theme=theme)

    Apply this theme to area plotting.

    >>> mapdl.aplot(theme=theme)

    """

    def __init__(self):
        """Initialize the theme."""
        super().__init__()

        self.name = "PyMAPDL-Theme"
        self.title = "PyMAPDL"
        self.background = "paraview"
        self.interactive = True

        if _HAS_MATPLOTLIB:
            self.cmap = PyMAPDL_cmap

        self.font.size = 18
        self.font.title_size = 18
        self.font.label_size = 18
        self.font.color = "black"
        self.font.family = "arial"

        self.axes.x_color = "tomato"
        self.axes.y_color = "seagreen"
        self.axes.z_color = "blue"

        self.show_edges = False
        self.color = "lightblue"
        self.outline_color = "black"
        self.edge_color = "black"

        self.color_cycler = get_cycler(MAPDL_colorbar.tolist())
        self.render_points_as_spheres = True


def _apply_default_theme():
    import pyvista as pv

    pv.global_theme = MapdlTheme()



# Copyright (C) 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class Listing:
    def extrem(self, nvar1="", nvar2="", ninc="", **kwargs):
        """Lists the extreme values for variables.

        APDL Command: EXTREM

        Parameters
        ----------
        nvar1, nvar2, ninc
            List extremes for variables NVAR1 through NVAR2 in steps of NINC.
            Variable range defaults to its maximum. NINC defaults to 1.

        Notes
        -----
        Lists the extreme values (and the corresponding times) for stored and
        calculated variables. Extremes for stored variables are automatically
        listed as they are stored. Only the real part of a complex number is
        used. Extreme values may also be assigned to parameters [``*GET``].
        """
        command = f"EXTREM,{nvar1},{nvar2},{ninc}"
        return self.run(command, **kwargs)

    def lines(self, n="", **kwargs):
        """Specifies the length of a printed page.

        APDL Command: LINES

        Parameters
        ----------
        n
            Number of lines per page (defaults to 20).  (Minimum allowed = 11).

        Notes
        -----
        Specifies the length of a printed page (for use in reports, etc.).
        """
        command = f"LINES,{n}"
        return self.run(command, **kwargs)

    def nprint(self, n="", **kwargs):
        """Defines which time points stored are to be listed.

        APDL Command: NPRINT

        Parameters
        ----------
        n
            List data associated with every N time (or frequency) point(s),
            beginning with the first point stored (defaults to 1).

        Notes
        -----
        Defines which time (or frequency) points within the range stored are to
        be listed.
        """
        command = f"NPRINT,{n}"
        return self.run(command, **kwargs)

    def prcplx(self, key="", **kwargs):
        """Defines the output form for complex variables.

        APDL Command: PRCPLX

        Parameters
        ----------
        key
            Output form key:

            0 - Real and imaginary parts.

            1 - Amplitude and phase angle.  Stored real and imaginary data are converted to
                amplitude and phase angle upon output.  Data remain stored as
                real and imaginary parts.

        Notes
        -----
        Defines the output form for complex variables.  Used only with harmonic
        analyses (ANTYPE,HARMIC).

        All results data are stored in the form of real and imaginary
        components and converted to amplitude and/or phase angle as specified
        via the PRCPLX command. The conversion is not  valid for derived
        results (such as principal stress/strain, equivalent stress/strain and
        USUM).
        """
        command = f"PRCPLX,{key}"
        return self.run(command, **kwargs)

    def prtime(self, tmin="", tmax="", **kwargs):
        """Defines the time range for which data are to be listed.

        APDL Command: PRTIME

        Parameters
        ----------
        tmin
            Minimum time (defaults to the first point stored).

        tmax
            Maximum time (defaults to the last point stored).

        Notes
        -----
        Defines the time (or frequency) range (within the range stored) for
        which data are to be listed.
        """
        command = f"PRTIME,{tmin},{tmax}"
        return self.run(command, **kwargs)

    def prvar(
        self,
        nvar1="",
        nvar2="",
        nvar3="",
        nvar4="",
        nvar5="",
        nvar6="",
        **kwargs,
    ):
        """Lists variables vs. time (or frequency).

        APDL Command: PRVAR

        Parameters
        ----------
        nvar1, nvar2, nvar3, . . . , nvar6
            Variables to be displayed, defined either by the reference number
            or a unique thirty-two character name. If duplicate names are used
            the command will print the data for the lowest-numbered variable
            with that name.

        Notes
        -----
        Lists variables vs. time (or frequency).  Up to six variables may be
        listed across the line. Time column output format can be changed using
        the /FORMAT command arguments Ftype, NWIDTH, and DSIGNF.
        """
        command = f"PRVAR,{nvar1},{nvar2},{nvar3},{nvar4},{nvar5},{nvar6}"
        return self.run(command, **kwargs)



# Copyright (C) 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Optional, Union

from ansys.mapdl.core._commands import parse
from ansys.mapdl.core.mapdl_types import MapdlInt


class Volumes:
    def extopt(
        self,
        lab: str = "",
        val1: Union[str, int] = "",
        val2: Union[str, int] = "",
        val3: MapdlInt = "",
        val4: MapdlInt = "",
        **kwargs,
    ) -> Optional[str]:
        """Controls options relating to the generation of volume elements from area elements.

        APDL Command: EXTOPT

        Parameters
        ----------
        lab
            Label identifying the control option. The meanings of
            Val1, Val2, and Val3 will vary depending on Lab.

            ON - Sets carryover of the material attributes, real
                 constant attributes, and element coordinate system
                 attributes of the pattern area elements to the
                 generated volume elements.  Sets the pattern
                 area mesh to clear when volume generations are done.
                 Val1, Val2, and Val3 are ignored.

            OFF - Removes all settings associated with this command.
                  Val1, Val2, and Val3 are ignored.

            STAT - Shows all settings associated with this command.
                   Val1, Val2, Val3, and Val4 are ignored.

            ATTR - Sets carryover of particular pattern area attributes
                   (materials, real constants, and element coordinate
                   systems) of the pattern area elements to the
                   generated volume elements. (See 2.) Val1 can be:

                0 - Sets volume elements to use
                    current MAT command settings.

                1 - Sets volume elements to use material
                    attributes of the pattern area elements.

            Val2 can be:

                0 - Sets volume elements to use current REAL
                    command settings.

                1 - Sets volume elements to use real constant attributes
                    of the pattern area elements.

            Val3 can be:

                0 - Sets volume elements to use current ESYS command
                    settings.

                1 - Sets volume elements to use element coordinate
                    system attributes of the pattern area elements.

            Val4 can be:

                0 - Sets volume elements to use current SECNUM command
                    settings.

                1 - Sets volume elements to use section attributes of
                    the pattern area elements.

            ESIZE - Val1 sets the number of element divisions in the
                    direction of volume generation or volume sweep.
                    For VDRAG and VSWEEP, Val1 is overridden by the
                    LESIZE command NDIV setting. Val2 sets the spacing
                    ratio (bias) in the direction of volume generation
                    or volume sweep. If positive, Val2 is the nominal
                    ratio of last division size to first division size
                    (if > 1.0, sizes increase, if < 1.0, sizes
                    decrease). If negative, Val2 is the nominal ratio of
                    center division(s) size to end divisions size. Ratio
                    defaults to 1.0 (uniform spacing).
                    Val3 and Val4 are ignored.

            ACLEAR - Sets clearing of pattern area mesh.
                     (See 3.) Val1 can be:

                0 - Sets pattern area to remain meshed when volume
                    generation is done.

                1 - Sets pattern area mesh to clear when volume
                    generation is done. Val2, Val3, and Val4 are
                    ignored.

            VSWE - Indicates that volume sweeping options will be set
                   using Val1 and Val2. Settings specified with EXTOPT,
                   VSWE will be used the next time the VSWEEP command
                   is invoked. If Lab = VSWE, Val1 becomes a label.
                   Val1 can be:

            AUTO - Indicates whether you will be prompted for the source
                   and target used by VSWEEP or if VSWE should
                   automatically determine the source and target.
                   If Val1 = AUTO, Val2 is ON by default. VSWE will
                   automatically determine the source and target for
                   VSWEEP. You will be allowed to pick more than one
                   volume for sweeping. When Val2 = OFF, the user will
                   be prompted for the source and target for VSWEEP.
                   You will only be allowed to pick one volume for
                   sweeping.

            TETS - Indicates whether VSWEEP will tet mesh non-sweepable
                   volumes or leave them unmeshed. If Val1 = TETS,
                   Val2 is OFF by default. Non-sweepable volumes will be
                   left unmeshed. When Val2 = ON, the non-sweepable
                   volumes will be tet meshed if the assigned element
                   type supports tet shaped elements.

        val1, val2, val3, val4
            Additional input values as described under each option for
            Lab.

        Notes
        -----
        EXTOPT controls options relating to the generation of volume
        elements from pattern area elements using the VEXT, VROTAT,
        VOFFST, VDRAG, and VSWEEP commands.  (When using VSWEEP,
        the pattern area is referred to as the source area.)

        Enables carryover of the attributes  of the pattern area
        elements to the generated volume elements when you are using
        VEXT, VROTAT, VOFFST, or VDRAG. (When using VSWEEP, since the
        volume already exists, use the VATT command to assign attributes
        before sweeping.)

        When you are using VEXT, VROTAT, VOFFST, or VDRAG, enables
        clearing of the pattern area mesh when volume generations are
        done. (When you are using VSWEEP, if selected, the area meshes
        on the pattern (source), target, and/or side areas clear when
        volume sweeping is done.)

        Neither EXTOPT,VSWE,AUTO nor EXTOPT,VSWE,TETS will be affected
        by EXTOPT,ON or EXTOPT, OFF.
        """
        command = f"EXTOPT,{lab},{val1},{val2},{val3},{val4}"
        return self.run(command, **kwargs)

    def v(
        self, p1="", p2="", p3="", p4="", p5="", p6="", p7="", p8="", **kwargs
    ) -> int:
        """Define a volume through keypoints.

        APDL Command: V

        Parameters
        ----------
        p1 : int, optional
            Keypoint defining starting corner of volume.

        p2 : int, optional
            Keypoint defining second corner of volume.

        p3 : int, optional
            Keypoint defining third corner of volume.

        p4 : int, optional
            Keypoint defining fourth corner of volume.

        p5 : int, optional
            Keypoint defining fifth corner of volume.

        p6 : int, optional
            Keypoint defining sixth corner of volume.

        p7 : int, optional
            Keypoint defining seventh corner of volume.

        p8 : int, optional
            Keypoint defining eighth corner of volume.

        Returns
        -------
        int
            Volume number of the generated volume.

        Notes
        -----
        Defines a volume (and its corresponding lines and areas)
        through eight (or fewer) existing keypoints.  Keypoints must
        be input in a continuous order.  The order of the keypoints
        should be around the bottom and then the top.  Missing lines
        are generated "straight" in the active coordinate system and
        assigned the lowest available numbers [NUMSTR].  Missing areas
        are generated and assigned the lowest available numbers.

        Solid modeling in a toroidal coordinate system is not recommended.

        Certain faces may be condensed to a line or point by repeating
        keypoints.   For example, use V,P1,P2,P3,P3,P5,P6,P7,P7   for a
        triangular prism or V,P1,P2,P3,P3,P5,P5,P5,P5  for a tetrahedron.

        Using keypoints to produce partial sections in CSYS = 2 can generate
        anomalies; check the resulting volumes carefully.

        Examples
        --------
        Create a simple cube volume.

        >>> k0 = mapdl.k("", 0, 0, 0)
        >>> k1 = mapdl.k("", 1, 0, 0)
        >>> k2 = mapdl.k("", 1, 1, 0)
        >>> k3 = mapdl.k("", 0, 1, 0)
        >>> k4 = mapdl.k("", 0, 0, 1)
        >>> k5 = mapdl.k("", 1, 0, 1)
        >>> k6 = mapdl.k("", 1, 1, 1)
        >>> k7 = mapdl.k("", 0, 1, 1)
        >>> v0 = mapdl.v(k0, k1, k2, k3, k4, k5, k6, k7)
        >>> v0
        1

        Create a triangular prism

        >>> k0 = mapdl.k("", 0, 0, 0)
        >>> k1 = mapdl.k("", 1, 0, 0)
        >>> k2 = mapdl.k("", 1, 1, 0)
        >>> k3 = mapdl.k("", 0, 1, 0)
        >>> k4 = mapdl.k("", 0, 0, 1)
        >>> k5 = mapdl.k("", 1, 0, 1)
        >>> k6 = mapdl.k("", 1, 1, 1)
        >>> k7 = mapdl.k("", 0, 1, 1)
        >>> v1 = mapdl.v(k0, k1, k2, k2, k4, k5, k6, k6)
        >>> v1
        2

        Create a tetrahedron

        >>> k0 = mapdl.k("", 0, 0, 0)
        >>> k1 = mapdl.k("", 1, 0, 0)
        >>> k2 = mapdl.k("", 1, 1, 0)
        >>> k3 = mapdl.k("", 0, 0, 1)
        >>> v2 = mapdl.v(k0, k1, k2, k2, k3, k3, k3, k3)
        >>> v2
        3

        """
        command = f"V,{p1},{p2},{p3},{p4},{p5},{p6},{p7},{p8}"
        return parse.parse_v(self.run(command, **kwargs))

    def va(
        self,
        a1="",
        a2="",
        a3="",
        a4="",
        a5="",
        a6="",
        a7="",
        a8="",
        a9="",
        a10="",
        **kwargs,
    ) -> int:
        """Generate a volume bounded by existing areas.

        APDL Command: VA

        Parameters
        ----------
        a1, a2, a3, . . . , a10
            List of areas defining volume.  The minimum number of
            areas is 4.  If A1 = ALL, use all selected [ASEL] areas
            and ignore A2 to A10.  A component name may also be
            substituted for A1.

        Returns
        -------
        int
            Volume number of the volume.

        Notes
        -----
        This command conveniently allows generating volumes from
        regions having more than eight keypoints (which is not allowed
        with the V command).  Areas may be input in any order.  The
        exterior surface of a VA volume must be continuous, but holes
        may pass completely through it.

        Examples
        --------
        Create a simple tetrahedral bounded by 4 areas.

        >>> k0 = mapdl.k('', -1, 0, 0)
        >>> k1 = mapdl.k('', 1, 0,  0)
        >>> k2 = mapdl.k('', 1, 1, 0)
        >>> k3 = mapdl.k('', 1, 0.5, 1)
        >>> a0 = mapdl.a(k0, k1, k2)
        >>> a1 = mapdl.a(k0, k1, k3)
        >>> a2 = mapdl.a(k1, k2, k3)
        >>> a3 = mapdl.a(k0, k2, k3)
        >>> vnum = mapdl.va(a0, a1, a2, a3)
        >>> vnum
        1

        """
        command = f"VA,{a1},{a2},{a3},{a4},{a5},{a6},{a7},{a8},{a9},{a10}"
        return parse.parse_v(self.run(command, **kwargs))

    def vdele(self, nv1="", nv2="", ninc="", kswp="", **kwargs):
        """Deletes unmeshed volumes.

        APDL Command: VDELE

        Parameters
        ----------
        nv1, nv2, ninc
            Delete volumes from NV1 to NV2 (defaults to NV1) in steps of NINC
            (defaults to 1).  If NV1 = ALL, NV2 and NINC are ignored and all
            selected volumes [VSEL] are deleted.  If NV1 = P, graphical picking
            is enabled and all remaining command fields are ignored (valid only
            in the GUI).  A component name may also be substituted for NV1 (NV2
            and NINC are ignored).

        kswp
            Specifies whether keypoints, lines, and areas are also deleted:

            0 - Delete volumes only (default).

            1 - Delete volumes, as well as keypoints, lines, and areas attached to the
                specified volumes but not shared by other volumes.
        """
        command = f"VDELE,{nv1},{nv2},{ninc},{kswp}"
        return self.run(command, **kwargs)

    def vdgl(self, nv1="", nv2="", ninc="", **kwargs):
        """Lists keypoints of a volume that lie on a parametric degeneracy.

        APDL Command: VDGL

        Parameters
        ----------
        nv1, nv2, ninc
            List keypoints that lie on a parametric degeneracy on volumes from
            NV1 to NV2 (defaults to NV1) in steps of NINC (defaults to 1).  If
            NV1 = ALL (default), NV2 and NINC will be ignored and keypoints on
            all selected volumes [VSEL] will be listed.  If NV1 = P, graphical
            picking is enabled and all remaining command fields are ignored
            (valid only in the GUI).  You may also substitute a component name
            for NV1 (ignore NV2 and NINC).

        Notes
        -----
        See the Modeling and Meshing Guide for details about parametric
        degeneracies.

        This command is valid in any processor.
        """
        command = f"VDGL,{nv1},{nv2},{ninc}"
        return self.run(command, **kwargs)

    def vdrag(
        self,
        na1="",
        na2="",
        na3="",
        na4="",
        na5="",
        na6="",
        nlp1="",
        nlp2="",
        nlp3="",
        nlp4="",
        nlp5="",
        nlp6="",
        **kwargs,
    ) -> str:
        """Generate volumes by dragging an area pattern along a path.

        APDL Command: VDRAG

        Parameters
        ----------
        na1, na2, na3, . . . , na6
            List of areas in the pattern to be dragged (6 maximum if
            using keyboard entry).  If NA1 = ALL, all selected areas
            will be swept along the path.  A component name may also
            be substituted for NA1.

        nlp1, nlp2, nlp3, . . . , nlp6
            List of lines defining the path along which the pattern is
            to be dragged (6 maximum if using keyboard entry).  Must
            be a continuous set of lines.  To be continuous, adjacent
            lines must share the connecting keypoint (the end keypoint
            of one line must also be first keypoint of the next line).

        Returns
        -------
        str
            MAPDL command output.

        Notes
        -----
        Generates volumes (and their corresponding keypoints, lines,
        and areas) by sweeping a given area pattern along a
        characteristic drag path.  If the drag path consists of
        multiple lines, the drag direction is determined by the
        sequence in which the path lines are input (NLP1, NLP2, etc.).
        If the drag path is a single line (NLP1), the drag direction
        is from the keypoint on the drag line that is closest to the
        first keypoint of the given area pattern to the other end of
        the drag line.

        The magnitude of the vector between the keypoints of the given
        pattern and the first path keypoint remains constant for all
        generated keypoint patterns and the path keypoints.  The
        direction of the vector relative to the path slope also
        remains constant so that patterns may be swept around curves.
        Lines are generated with the same shapes as the given pattern
        and the path lines.

        Keypoint, line, area, and volume numbers are automatically
        assigned (beginning with the lowest available values
        [NUMSTR]).  Adjacent lines use a common keypoint, adjacent
        areas use a common line, and adjacent volumes use a common
        area.  For best results, the entities to be dragged should be
        orthogonal to the start of the drag path.  Drag operations
        that produce an error message may create some of the desired
        entities prior to terminating.

        If element attributes have been associated with the input area
        via the AATT command, the opposite area generated by the VDRAG
        operation will also have those attributes (i.e., the element
        attributes from the input area are copied to the opposite
        area).  Note that only the area opposite the input area will
        have the same attributes as the input area; the areas adjacent
        to the input area will not.

        If the input areas are meshed or belong to a meshed volume,
        the area(s) can be extruded to a 3-D mesh.  Note that the NDIV
        argument of the ESIZE command should be set before extruding
        the meshed areas.  Alternatively, mesh divisions can be
        specified directly on the drag line(s) (LESIZE).  See the
        Modeling and Meshing Guide for more information.

        You can use the VDRAG command to generate 3-D interface
        element meshes for elements INTER194 and INTER195. When
        generating interface element meshes using VDRAG, you must
        specify the line divisions to generate one interface element
        directly on the drag line using the LESIZE command.  The
        source area to be extruded becomes the bottom surface of the
        interface element. Interface elements must be extruded in what
        will become the element's local x direction, that is, bottom
        to top.

        Examples
        --------
        Create a square with a hole in it and drag it along an arc.

        >>> anum0 = mapdl.blc4(0, 0, 1, 1)
        >>> anum1 = mapdl.blc4(0.25, 0.25, 0.5, 0.5)
        >>> aout = mapdl.asba(anum0, anum1)
        >>> k0 = mapdl.k("", 0, 0, 0)
        >>> k1 = mapdl.k("", 1, 0, 1)
        >>> k2 = mapdl.k("", 1, 0, 0)
        >>> l0 = mapdl.larc(k0, k1, k2, 2)
        >>> output = mapdl.vdrag(aout, nlp1=l0)
        >>> print(output)
        DRAG AREAS
          3,
        ALONG LINES
          9

        """
        command = f"VDRAG,{na1},{na2},{na3},{na4},{na5},{na6},{nlp1},{nlp2},{nlp3},{nlp4},{nlp5},{nlp6}"
        return self.run(command, **kwargs)

    def vext(
        self,
        na1="",
        na2="",
        ninc="",
        dx="",
        dy="",
        dz="",
        rx="",
        ry="",
        rz="",
        **kwargs,
    ) -> str:
        """Generate additional volumes by extruding areas.

        APDL Command: VEXT

        Parameters
        ----------
        na1, na2, ninc
            Set of areas (NA1 to NA2 in steps of NINC) that defines
            the pattern to be extruded.  NA2 defaults to NA1, NINC
            defaults to 1.  If NA1 = ALL, NA2 and NINC are ignored and
            the pattern is defined by all selected areas.  A component
            name may also be substituted for NA1 (NA2 and NINC are
            ignored).

        dx, dy, dz
            Increments to be applied to the X, Y, and Z keypoint
            coordinates in the active coordinate system (DR, Dθ, DZ
            for cylindrical; DR, Dθ, DΦ for spherical).

        rx, ry, rz
            Scale factors to be applied to the X, Y, and Z keypoint
            coordinates in the active coordinate system (RR, Rθ, RZ
            for cylindrical; RR, Rθ, RΦ for spherical).  Note that the
            Rθ and RΦ scale factors are interpreted as angular
            offsets.  For example, if CSYS = 1, RX, RY, RZ input of
            (1.5,10,3) would scale the specified keypoints 1.5 times
            in the radial and 3 times in the Z direction, while adding
            an offset of 10 degrees to the keypoints.  Zero, blank, or
            negative scale factor values are assumed to be 1.0.  Zero
            or blank angular offsets have no effect.

        Returns
        -------
        str
            MAPDL command output.

        Examples
        --------
        Create a basic cylinder by extruding a circle.

        >>> k0 = mapdl.k("", 0, 0, 0)
        >>> k1 = mapdl.k("", 0, 0, 1)
        >>> k2 = mapdl.k("", 0, 0, 0.5)
        >>> carc0 = mapdl.circle(k0, 1, k1)
        >>> a0 = mapdl.al(*carc0)
        >>> mapdl.vext(a0, dz=4)

        Create a tapered cylinder.

        >>> mapdl.vdele('all')
        >>> mapdl.vext(a0, dz=4, rx=0.3, ry=0.3, rz=1)

        """
        command = f"VEXT,{na1},{na2},{ninc},{dx},{dy},{dz},{rx},{ry},{rz}"
        return self.run(command, **kwargs)

    def vgen(
        self,
        itime="",
        nv1="",
        nv2="",
        ninc="",
        dx="",
        dy="",
        dz="",
        kinc="",
        noelem="",
        imove="",
        **kwargs,
    ):
        """Generates additional volumes from a pattern of volumes.

        APDL Command: VGEN

        Generates additional volumes (and their corresponding
        keypoints, lines, and areas) by extruding and scaling a
        pattern of areas in the active coordinate system.

        If element attributes have been associated with the input area
        via the AATT command, the opposite area generated by the VEXT
        operation will also have those attributes (i.e., the element
        attributes from the input area are copied to the opposite
        area).  Note that only the area opposite the input area will
        have the same attributes as the input area; the areas adjacent
        to the input area will not.

        If the areas are meshed or belong to meshed volumes, a 3-D
        mesh can be extruded with this command.  Note that the NDIV
        argument on the ESIZE command should be set before extruding
        the meshed areas.

        Scaling of the input areas, if specified, is performed first,
        followed by the extrusion.

        In a non-Cartesian coordinate system, the VEXT command locates
        the end face of the volume based on the active coordinate
        system.  However, the extrusion is made along a straight line
        between the end faces.  Note that solid modeling in a toroidal
        coordinate system is not recommended.

        .. warning::
           Use of the VEXT command can produce unexpected results when
           operating in a non-Cartesian coordinate system.  For a
           detailed description of the possible problems that may
           occur, see Solid Modeling in the Modeling and Meshing
           Guide.

        Parameters
        ----------
        itime
            Do this generation operation a total of ITIMEs, incrementing all
            keypoints in the given pattern automatically (or by KINC) each time
            after the first.  ITIME must be > 1 for generation to occur.

        nv1, nv2, ninc
            Generate volumes from pattern beginning with NV1 to NV2 (defaults
            to NV1) in steps of NINC (defaults to 1).  If NV1 = ALL, NV2 and
            NINC are ignored and the pattern is all selected volumes [VSEL].
            If NV1 = P, graphical picking is enabled and all remaining command
            fields are ignored (valid only in the GUI).  A component name may
            also be substituted for NV1 (NV2 and NINC are ignored).

        dx, dy, dz
            Keypoint location increments in the active coordinate system (--,
            Dθ, DZ for cylindrical, --, Dθ, -- for spherical).

        kinc
            Keypoint increment between generated sets.  If zero, the lowest
            available keypoint numbers are assigned [NUMSTR].

        noelem
            Specifies if elements and nodes are also to be generated:

            0 - Generate nodes and elements associated with the original volumes, if they
                exist.

            1 - Do not generate nodes and elements.

        imove
            Specifies whether to redefine the existing volumes:

            0 - Generate additional volumes as requested with the ITIME argument.

            1 - Move original volumes to new position retaining the same keypoint line, and
                area numbers (ITIME, KINC, and NOELEM are ignored).
                Corresponding meshed items are also moved if not needed at
                their original position.

        Notes
        -----
        Generates additional volumes (and their corresponding keypoints, lines,
        areas and mesh) from a given volume pattern.  The MAT, TYPE, REAL, and
        ESYS attributes are based upon the volumes in the pattern and not upon
        the current settings of the pointers.  End slopes of the generated
        lines remain the same (in the active coordinate system) as those of the
        given pattern.  For example, radial slopes remain radial, etc.
        Generations which produce volumes of a size or shape different from the
        pattern (i.e., radial generations in cylindrical systems, radial and
        phi generations in spherical systems, and theta generations in
        elliptical systems) are not allowed.  Note that solid modeling in a
        toroidal coordinate system is not recommended.  Volume, area, and line
        numbers are automatically assigned (beginning with the lowest available
        values [NUMSTR]).
        """
        command = (
            f"VGEN,{itime},{nv1},{nv2},{ninc},{dx},{dy},{dz},{kinc},{noelem},{imove}"
        )
        return self.run(command, **kwargs)

    def vlist(self, nv1="", nv2="", ninc="", **kwargs):
        """Lists the defined volumes.

        APDL Command: VLIST

        Parameters
        ----------
        nv1, nv2, ninc
            List volumes from NV1 to NV2 (defaults to NV1) in steps of NINC
            (defaults to 1).  If NV1 = ALL (default), NV2 and NINC are ignored
            and all selected volumes [VSEL] are listed.  If NV1 = P, graphical
            picking is enabled and all remaining command fields are ignored
            (valid only in the GUI).  A component name may also be substituted
            for NV1 (NV2 and NINC are ignored).

        Notes
        -----
        An attribute (TYPE, MAT, REAL, or ESYS) listed as a zero is unassigned;
        one listed as a positive value indicates that the attribute was
        assigned with the VATT command (and will not be reset to zero if the
        mesh is cleared); one listed as a negative value indicates that the
        attribute was assigned using the attribute pointer [TYPE, MAT, REAL, or
        ESYS] that was active during meshing (and will be reset to zero if the
        mesh is cleared).  A "-1" in the "nodes" column indicates that the
        volume has been meshed but there are no interior nodes.  The volume
        size is listed only if a VSUM command has been performed on the volume.
        Volume orientation attributes (KZ1 and KZ2) are listed only if a
        VEORIENT command was previously used to define an orientation for the
        volume.

        This command is valid in any processor.
        """
        command = f"VLIST,{nv1},{nv2},{ninc}"
        return self.run(command, **kwargs)

    def vlscale(
        self,
        nv1="",
        nv2="",
        ninc="",
        rx="",
        ry="",
        rz="",
        kinc="",
        noelem="",
        imove="",
        **kwargs,
    ):
        """Generates a scaled set of volumes from a pattern of volumes.

        APDL Command: VLSCALE

        Parameters
        ----------
        nv1, nv2, ninc
            Set of volumes (NV1 to NV2 in steps of NINC) that defines the
            pattern to be scaled.  NV2 defaults to NV1, NINC defaults to 1.  If
            NV1 = ALL, NV2 and NINC are ignored and the pattern is defined by
            all selected volumes.  If NV1 = P, graphical picking is enabled and
            all remaining command fields are ignored (valid only in the GUI).
            A component name may also be substituted for NV1 (NV2 and NINC are
            ignored).

        rx, ry, rz
            Scale factors to be applied to the X, Y, and Z keypoint coordinates
            in active coordinate system (RR, Rθ, RZ for cylindrical; RR, Rθ, RΦ
            for spherical).  Note that the Rθ and RΦ scale factors are
            interpreted as angular offsets.  For example, if CSYS = 1, RX, RY,
            RZ input of (1.5,10,3) would scale the specified keypoints 1.5
            times in the radial and 3 times in the Z direction, while adding an
            offset of 10 degrees to the keypoints.  Zero, blank, or negative
            scale factor values are assumed to be 1.0.  Zero or blank angular
            offsets have no effect.

        kinc
            Increment to be applied to keypoint numbers for generated set.  If
            zero, the lowest available keypoint numbers will be assigned
            [NUMSTR].

        noelem
            Specifies whether nodes and elements are also to be generated:

            0 - Nodes and elements associated with the original volumes will be generated
                (scaled) if they exist.

            1 - Nodes and elements will not be generated.

        imove
            Specifies whether volumes will be moved or newly defined:

            0 - Additional volumes will be generated.

            1 - Original volumes will be moved to new position (KINC and NOELEM are ignored).
                Use only if the old volumes are no longer needed at their
                original positions.  Corresponding meshed items are also moved
                if not needed at their original position.

        Notes
        -----
        Generates a scaled set of volumes (and their corresponding keypoints,
        lines, areas,  and mesh) from a pattern of volumes.  The MAT, TYPE,
        REAL, and ESYS attributes are based on the volumes in the pattern and
        not the current settings.  Scaling is done in the active coordinate
        system.  Volumes in the pattern could have been generated in any
        coordinate system.  However, solid modeling in a toroidal coordinate
        system is not recommended.
        """
        command = f"VLSCALE,{nv1},{nv2},{ninc},{rx},{ry},{rz},{kinc},{noelem},{imove}"
        return self.run(command, **kwargs)

    def voffst(self, narea="", dist="", kinc="", **kwargs):
        """Generates a volume, offset from a given area.

        APDL Command: VOFFST

        Parameters
        ----------
        narea
            Area from which generated volume is to be offset.  If NAREA = P,
            graphical picking is enabled and all remaining command fields are
            ignored (valid only in the GUI).

        dist
            Distance normal to given area at which keypoints for generated
            volume are to be located.  Positive normal is determined from the
            right-hand rule keypoint order.

        kinc
            Increment to be applied to the keypoint numbers between sets.  If
            zero, keypoint numbers will be automatically assigned beginning
            with the lowest available value [NUMSTR].

        Notes
        -----
        Generates a volume (and its corresponding keypoints, lines, and areas)
        by offsetting from an area.  The direction of the offset varies with
        the given area normal.  End slopes of the generated lines remain the
        same as those of the given pattern.

        If element attributes have been associated with the input area via the
        AATT command, the opposite area generated by the VOFFST operation will
        also have those attributes (i.e., the element attributes from the input
        area are copied to the opposite area).  Note that only the area
        opposite the input area will have the same attributes as the input
        area; the areas adjacent to the input area will not.

        If the areas are meshed or belong to meshed volumes, a 3-D mesh can be
        extruded with this command.  Note that the NDIV argument on the ESIZE
        command should be set before extruding the meshed areas.
        """
        command = f"VOFFST,{narea},{dist},{kinc}"
        return self.run(command, **kwargs)

    def vplot(self, nv1="", nv2="", ninc="", degen="", scale="", **kwargs):
        """Displays the selected volumes.

        APDL Command: VPLOT

        Parameters
        ----------
        nv1, nv2, ninc
            Display volumes from NV1 to NV2 (defaults to NV1) in steps of NINC
            (defaults to 1).  If NV1 = ALL (default), NV2 and NINC are ignored
            and all selected volumes [VSEL] are displayed.

        degen
            Degeneracy marker:

            (blank) - No degeneracy marker is used (default).

            DEGE - A red star is placed on keypoints at degeneracies (see the Modeling and Meshing
                   Guide).  Not available if /FACET,WIRE is set.

        scale
            Scale factor for the size of the degeneracy-marker star.  The scale
            is the size in window space (-1 to 1 in both directions) (defaults
            to .075).

        Notes
        -----
        Displays selected volumes.  (Only volumes having areas within the
        selected area set [ASEL] will be plotted.)  With PowerGraphics on
        [/GRAPHICS,POWER], VPLOT will display only the currently selected
        areas. This command is also a utility command, valid anywhere.  The
        degree of tessellation used to plot the volumes is set through the
        /FACET command.
        """
        command = f"VPLOT,{nv1},{nv2},{ninc},{degen},{scale}"
        return self.run(command, **kwargs)

    def vrotat(
        self,
        na1="",
        na2="",
        na3="",
        na4="",
        na5="",
        na6="",
        pax1="",
        pax2="",
        arc="",
        nseg="",
        **kwargs,
    ) -> str:
        """Generate cylindrical volumes by rotating an area pattern about an axis.

        APDL Command: VROTAT

        Generates cylindrical volumes (and their corresponding
        keypoints, lines, and areas) by rotating an area pattern (and
        its associated line and keypoint patterns) about an axis.
        Keypoint patterns are generated at regular angular locations
        (based on a maximum spacing of 90 degrees).  Line patterns are
        generated at the keypoint patterns.  Arc lines are also
        generated to connect the keypoints circumferentially.
        Keypoint, line, area, and volume numbers are automatically
        assigned (beginning with the lowest available values).
        Adjacent lines use a common keypoint, adjacent areas use a
        common line, and adjacent volumes use a common area.

        To generate a single volume with an arc greater than 180 degrees,
        NSEG must be greater than or equal to 2.

        If element attributes have been associated with the input area
        via the AATT command, the opposite area generated by the
        VROTAT operation will also have those attributes (i.e., the
        element attributes from the input area are copied to the
        opposite area).  Note that only the area opposite the input
        area will have the same attributes as the input area; the
        areas adjacent to the input area will not.

        If the given areas are meshed or belong to meshed volumes, the
        2-D mesh can be rotated (extruded) to a 3-D mesh. See the
        Modeling and Meshing Guide for more information.  Note that
        the NDIV argument on the ESIZE command should be set before
        extruding the meshed areas.

        Parameters
        ----------
        na1, na2, na3, . . . , na6
            List of areas in the pattern to be rotated (6 maximum if
            using keyboard entry).  Areas must lie to one side of, and
            in the plane of, the axis of rotation.  If NA1 = ALL,
            all selected areas will define the pattern to be rotated.
            A component name may also be substituted for NA1.

        pax1, pax2
            Keypoints defining the axis about which the area pattern
            is to be rotated.

        arc
            Arc length (in degrees).  Positive follows right-hand rule
            about PAX1-PAX2 vector.  Defaults to 360.

        nseg
            Number of volumes (8 maximum) around circumference.
            Defaults to minimum required for 90 degrees (maximum) arcs, i.e.,
            4 for 360 degrees, 3 for 270 degrees, etc.

        Returns
        -------
        str
            MAPDL command output.

        Examples
        --------
        Rotate a circle about the Z axis to create a hoop.

        First, create a circle offset from origin on the XZ plane.

        >>> hoop_radius = 10
        >>> hoop_thickness = 0.5
        >>> k0 = mapdl.k("", hoop_radius, 0, 0)
        >>> k1 = mapdl.k("", hoop_radius, 1, 0)
        >>> k2 = mapdl.k("", hoop_radius, 0, hoop_thickness)
        >>> carc0 = mapdl.circle(k0, 1, k1)
        >>> a0 = mapdl.al(*carc0)

        Create a hoop by rotating it about an axis defined by two
        keypoints.

        >>> k_axis0 = mapdl.k("", 0, 0, 0)
        >>> k_axis1 = mapdl.k("", 0, 0, 1)
        mapdl.vrotat(a0, pax1=k_axis0, pax2=k_axis1)

        """
        command = (
            f"VROTAT,{na1},{na2},{na3},{na4},{na5},{na6},{pax1},{pax2},{arc},{nseg}"
        )
        return self.run(command, **kwargs)

    def vsum(self, lab="", **kwargs):
        """Calculates and prints geometry statistics of the selected volumes.

        APDL Command: VSUM

        Parameters
        ----------
        lab
            Controls the degree of tessellation used in the calculation of area
            properties.  If LAB = DEFAULT, area calculations will use the
            degree of tessellation set through the /FACET command.  If LAB =
            FINE, area calculations are based on a finer tessellation.

        Notes
        -----
        Calculates and prints geometry statistics (volume, centroid location,
        moments of inertia, etc.) associated with the selected volumes.
        Geometry items are reported in the global Cartesian coordinate system.
        A unit density is assumed unless the volumes have a material
        association via the VATT command.  Items calculated by VSUM and later
        retrieved by a ``*GET`` or ``*VGET`` command are valid only if the model is not
        modified after the VSUM command is issued.

        Setting a finer degree of tessellation will provide area calculations
        with greater accuracy, especially for thin, hollow models.  However,
        using a finer degree of tessellation requires longer processing.

        For very thin volumes, such that the ratio of the minimum to the
        maximum dimension is less than 0.01, the VSUM command can provide
        erroneous volume information.  To ensure that such calculations are
        accurate, make certain that you subdivide such volumes so that the
        ratio of the minimum to the maximum is at least 0.05.
        """
        command = f"VSUM,{lab}"
        return self.run(command, **kwargs)

    def vsymm(
        self,
        ncomp="",
        nv1="",
        nv2="",
        ninc="",
        kinc="",
        noelem="",
        imove="",
        **kwargs,
    ) -> str:
        """Generate volumes from a volume pattern by symmetry reflection.

        APDL Command: VSYMM

        Generates a reflected set of volumes (and their corresponding
        keypoints, lines, areas and mesh) from a given volume pattern
        by a symmetry reflection (see analogous node symmetry command,
        NSYM).  The MAT, TYPE, REAL, and ESYS attributes are based
        upon the volumes in the pattern and not upon the current
        settings.  Reflection is done in the active coordinate system
        by changing a particular coordinate sign.  The active
        coordinate system must be a Cartesian system.  Volumes in the
        pattern may have been generated in any coordinate system.
        However, solid modeling in a toroidal coordinate system is not
        recommended.  Volumes are generated as described in the VGEN
        command.

        See the ESYM command for additional information about symmetry
        elements.

        Parameters
        ----------
        ncomp
            Symmetry key:

            - ``'X'`` : X symmetry (default).
            - ``'Y'`` : Y symmetry.
            - ``'Z'`` : Z symmetry.

        nv1, nv2, ninc
            Reflect volumes from pattern beginning with NV1 to NV2
            (defaults to NV1) in steps of NINC (defaults to 1).  If
            NV1 = ALL, NV2 and NINC are ignored and the pattern is all
            selected volumes [VSEL].  A component name may also be
            substituted for NV1 (NV2 and NINC are ignored).

        kinc
            Keypoint increment between sets.  If zero, the lowest
            available keypoint numbers are assigned [NUMSTR].

        noelem
            Specifies whether nodes and elements are also to be generated:

            0 - Generate nodes and elements associated with the
                original volumes, if they exist.

            1 - Do not generate nodes and elements.

        imove
            Specifies whether volumes will be moved or newly defined:

            0 - Generate additional volumes.

            1 - Move original volumes to new position retaining the
                same keypoint numbers (KINC and NOELEM are ignored).
                Corresponding meshed items are also moved if not
                needed at their original position.

        Returns
        -------
        str
            MAPDL command output.

        Examples
        --------
        Create four blocks by reflecting a single block across the X
        component and then the Y component.

        >>> vnum = mapdl.blc4(1, 1, 1, 1, depth=1)
        >>> mapdl.vsymm('X', vnum)
        >>> output = mapdl.vsymm('Y', 'ALL')
        >>> print(output)
        SYMMETRY TRANSFORMATION OF VOLUMES       USING COMPONENT  Y
           SET IS ALL SELECTED VOLUMES

        """
        command = f"VSYMM,{ncomp},{nv1},{nv2},{ninc},{kinc},{noelem},{imove}"
        return self.run(command, **kwargs)

    def vtran(
        self,
        kcnto="",
        nv1="",
        nv2="",
        ninc="",
        kinc="",
        noelem="",
        imove="",
        **kwargs,
    ):
        """Transfers a pattern of volumes to another coordinate system.

        APDL Command: VTRAN

        Parameters
        ----------
        kcnto
            Reference number of coordinate system where the pattern is to be
            transferred.  Transfer occurs from the active coordinate system.
            The coordinate system type and parameters of KCNTO must be the same
            as the active system.

        nv1, nv2, ninc
            Transfer volumes from pattern beginning with NV1 to NV2 (defaults
            to NV1) in steps of NINC (defaults to 1).  If NV1 = ALL, NV2 and
            NINC are ignored and the pattern is all selected volumes [VSEL].
            If NV1 = P, graphical picking is enabled and all remaining command
            fields are ignored (valid only in the GUI).  A component name may
            also be substituted for NV1 (NV2 and NINC are ignored).

        kinc
            Keypoint increment between sets.  If zero, the lowest available
            keypoint numbers are assigned [NUMSTR].

        noelem
            Specifies whether elements and nodes are also to be generated:

            0 - Generate nodes and elements associated with the original volumes, if they
                exist.

            1 - Do not generate nodes and elements.

        imove
            Specifies whether to redefine the existing volumes:

            0 - Generate additional volumes.

            1 - Move original volumes to new position retaining the same keypoint numbers (KINC
                and NOELEM are ignored).  Corresponding meshed items are also
                moved if not needed at their original position.

        Notes
        -----
        Transfers a pattern of volumes (and their corresponding keypoints,
        lines, areas and mesh) from one coordinate system to another (see
        analogous node transfer command, TRANSFER).  The MAT, TYPE, REAL, and
        ESYS attributes are based upon the volumes in the pattern and not upon
        the current settings.  Coordinate systems may be translated and rotated
        relative to each other.  Initial pattern may be generated in any
        coordinate system.  However, solid modeling in a toroidal coordinate
        system is not recommended.  Coordinate and slope values are interpreted
        in the active coordinate system and are transferred directly.  Volumes
        are generated as described in the VGEN command.
        """
        command = f"VTRAN,{kcnto},{nv1},{nv2},{ninc},{kinc},{noelem},{imove}"
        return self.run(command, **kwargs)



# Copyright (C) 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class RadiationMat:
    def emis(self, mat="", evalu="", **kwargs):
        """Specifies emissivity as a material property for the Radiation Matrix

        APDL Command: EMIS
        method.

        Parameters
        ----------
        mat
            Material number associated with this emissivity (500 maximum).
            Defaults to 1.

        evalu
            Emissivity for this material (0.0 < EVALU   1.0).  Enter a very
            small number for zero.

        Notes
        -----
        Specifies emissivity as a material property for the Radiation Matrix
        method. This material property can then be associated with each
        element.
        """
        command = f"EMIS,{mat},{evalu}"
        return self.run(command, **kwargs)

    def geom(self, k2d="", ndiv="", **kwargs):
        """Defines the geometry specifications for the radiation matrix

        APDL Command: GEOM
        calculation.

        Parameters
        ----------
        k2d
            Dimensionality key:

            0
                3-D geometry (default)

            1
                2-D geometry (plane or axisymmetric)

        ndiv
            Number of divisions in an axisymmetric model.  Used only with K2D =
            1.  Defaults to 0 (2-D plane).  The 2-D model is internally
            expanded to a 3-D model based on the number of divisions specified
            (6   NDIV   90).  For example, NDIV of 6 is internally represented
            by six 60° sections.
        """
        command = f"GEOM,{k2d},{ndiv}"
        return self.run(command, **kwargs)

    def mprint(self, key="", **kwargs):
        """Specifies that radiation matrices are to be printed.

        APDL Command: MPRINT

        Parameters
        ----------
        key
            Print key:

            0
                Do not print matrices.

            1
                Print matrices.

        Notes
        -----
        Specifies that the element and node radiation matrices are to be
        printed when the WRITE command is issued.  If KEY = 1, form factor
        information for each element will also be printed.
        """
        command = f"MPRINT,{key}"
        return self.run(command, **kwargs)

    def space(self, node="", **kwargs):
        """Defines a space node for radiation using the Radiation Matrix method.

        APDL Command: SPACE

        Parameters
        ----------
        node
            Node defined to be the space node.

        Notes
        -----
        A space node is required in an open system to account for radiation
        losses.

        If using SPACE with the ANSYS Multi-field solver (MFS),  you must
        capture this command in the command file using MFCMMAND. This step is
        necessary because at the end of each field computation, this command is
        unset.
        """
        command = f"SPACE,{node}"
        return self.run(command, **kwargs)

    def vtype(self, nohid="", nzone="", **kwargs):
        """Specifies the viewing procedure used to determine the form factors for

        APDL Command: VTYPE
        the Radiation Matrix method.

        Parameters
        ----------
        nohid
            Type of viewing procedure:

            0
                Hidden procedure.

            1
                Non-hidden (faster, but less general) procedure.

        nzone
            Number of sampling zones for the hidden procedure (100 maximum for
            3-D, 1000 maximum for 2-D).  Defaults to 20 for 3-D, 200 for 2-D.
            Number of points is ``2*NZONE`` for 2-D and ``2*NZONE*(NZONE+1)`` for 3-D.
        """
        command = f"VTYPE,{nohid},{nzone}"
        return self.run(command, **kwargs)

    def write(self, fname="", **kwargs):
        """Writes the radiation matrix file.

        APDL Command: WRITE

        Parameters
        ----------
        fname
            File name and directory path (248 characters maximum, including the
            characters needed for the directory path).  An unspecified
            directory path defaults to the working directory; in this case, you
            can use all 248 characters for the file name.

        Notes
        -----
        Writes radiation matrix file (File.SUB) for input to the substructure
        thermal "use" pass.  Subsequent WRITE operations to the same file
        overwrite the file.
        """
        command = f"WRITE,{fname}"
        return self.run(command, **kwargs)
